@@ -20,10 +20,9 @@
 from collections import defaultdict
 
 from dulwich.objects import (
-    Commit,
-    Tag,
     ZERO_SHA,
     )
+from dulwich.repo import check_ref_format
 
 from bzrlib import (
     branch,
@@ -34,6 +33,7 @@ from bzrlib import (
     revision,
     tag,
     transport,
+    urlutils,
     )
 from bzrlib.decorators import (
     needs_read_lock,
@@ -44,18 +44,18 @@ from bzrlib.revision import (
 from bzrlib.trace import (
     is_quiet,
     mutter,
+    warning,
     )
 
 from bzrlib.plugins.git.config import (
     GitBranchConfig,
+    GitBranchStack,
     )
 from bzrlib.plugins.git.errors import (
     NoPushSupport,
     NoSuchRef,
     )
 from bzrlib.plugins.git.refs import (
-    branch_name_to_ref,
-    extract_tags,
     is_tag,
     ref_to_branch_name,
     ref_to_tag_name,
@@ -74,7 +74,11 @@ class GitPullResult(branch.PullResult):
     def _lookup_revno(self, revid):
         assert isinstance(revid, str), "was %r" % revid
         # Try in source branch first, it'll be faster
-        return self.target_branch.revision_id_to_revno(revid)
+        self.target_branch.lock_read()
+        try:
+            return self.target_branch.revision_id_to_revno(revid)
+        finally:
+            self.target_branch.unlock()
 
     @property
     def old_revno(self):
@@ -92,37 +96,85 @@ class GitTags(tag.BasicTags):
         self.branch = branch
         self.repository = branch.repository
 
-    def get_refs(self):
-        raise NotImplementedError(self.get_refs)
+    def get_refs_container(self):
+        raise NotImplementedError(self.get_refs_container)
 
     def _iter_tag_refs(self, refs):
-        raise NotImplementedError(self._iter_tag_refs)
+        """Iterate over the tag refs.
+
+        :param refs: Refs dictionary (name -> git sha1)
+        :return: iterator over (name, peeled_sha1, unpeeled_sha1, bzr_revid)
+        """
+        for k, unpeeled in refs.as_dict().iteritems():
+            try:
+                tag_name = ref_to_tag_name(k)
+            except (ValueError, UnicodeDecodeError):
+                continue
+            peeled = refs.get_peeled(k)
+            if peeled is None:
+                peeled = self.repository.bzrdir._git.object_store.peel_sha(unpeeled).id
+            yield (tag_name, peeled, unpeeled,
+                   self.branch.lookup_foreign_revision_id(peeled))
+
+    def _merge_to_remote_git(self, target_repo, new_refs, overwrite=False):
+        updates = {}
+        conflicts = []
+        def get_changed_refs(old_refs):
+            ret = dict(old_refs)
+            for k, v in new_refs.iteritems():
+                if not is_tag(k):
+                    continue
+                name = ref_to_tag_name(k)
+                if old_refs.get(k) == v:
+                    pass
+                elif overwrite or not k in old_refs:
+                    ret[k] = v
+                    updates[name] = target_repo.lookup_foreign_revision_id(v)
+                else:
+                    conflicts.append((name, v, old_refs[k]))
+            return ret
+        target_repo.bzrdir.send_pack(get_changed_refs, lambda have, want: [])
+        return updates, conflicts
+
+    def _merge_to_local_git(self, target_repo, refs, overwrite=False):
+        conflicts = []
+        updates = {}
+        for k, unpeeled in refs.as_dict().iteritems():
+            if not is_tag(k):
+                continue
+            name = ref_to_tag_name(k)
+            peeled = self.repository.bzrdir.get_peeled(k)
+            if target_repo._git.refs.get(k) == unpeeled:
+                pass
+            elif overwrite or not k in target_repo._git.refs:
+                target_repo._git.refs[k] = unpeeled or peeled
+                updates[name] = target_repo.lookup_foreign_revision_id(peeled)
+            else:
+                conflicts.append((name, peeled, target_repo._git.refs[k]))
+        return updates, conflicts
 
     def _merge_to_git(self, to_tags, refs, overwrite=False):
         target_repo = to_tags.repository
-        conflicts = []
-        for k, v in refs.iteritems():
-            if not is_tag(k):
-                continue
-            if overwrite or not k in target_repo._git.refs:
-                target_repo._git.refs[k] = v
-            elif target_repo._git.refs[k] == v:
-                pass
-            else:
-                conflicts.append((ref_to_tag_name(k), v, target_repo.refs[k]))
-        return conflicts
+        if self.repository.has_same_location(target_repo):
+            return {}, []
+        if getattr(target_repo, "_git", None):
+            return self._merge_to_local_git(target_repo, refs, overwrite)
+        else:
+            return self._merge_to_remote_git(target_repo, refs, overwrite)
 
     def _merge_to_non_git(self, to_tags, refs, overwrite=False):
         unpeeled_map = defaultdict(set)
         conflicts = []
+        updates = {}
         result = dict(to_tags.get_tag_dict())
         for n, peeled, unpeeled, bzr_revid in self._iter_tag_refs(refs):
             if unpeeled is not None:
                 unpeeled_map[peeled].add(unpeeled)
-            if n not in result or overwrite:
-                result[n] = bzr_revid
-            elif result[n] == bzr_revid:
+            if result.get(n) == bzr_revid:
                 pass
+            elif n not in result or overwrite:
+                result[n] = bzr_revid
+                updates[n] = bzr_revid
             else:
                 conflicts.append((n, result[n], bzr_revid))
         to_tags._set_tag_dict(result)
@@ -130,15 +182,15 @@ class GitTags(tag.BasicTags):
             map_file = UnpeelMap.from_repository(to_tags.branch.repository)
             map_file.update(unpeeled_map)
             map_file.save_in_repository(to_tags.branch.repository)
-        return conflicts
+        return updates, conflicts
 
     def merge_to(self, to_tags, overwrite=False, ignore_master=False,
                  source_refs=None):
         """See Tags.merge_to."""
         if source_refs is None:
-            source_refs = self.get_refs()
+            source_refs = self.get_refs_container()
         if self == to_tags:
-            return
+            return {}, []
         if isinstance(to_tags, GitTags):
             return self._merge_to_git(to_tags, source_refs,
                                       overwrite=overwrite)
@@ -147,17 +199,20 @@ class GitTags(tag.BasicTags):
                 master = None
             else:
                 master = to_tags.branch.get_master_branch()
-            conflicts = self._merge_to_non_git(to_tags, source_refs,
+            updates, conflicts = self._merge_to_non_git(to_tags, source_refs,
                                               overwrite=overwrite)
             if master is not None:
-                conflicts += self.merge_to(master.tags, overwrite=overwrite,
+                extra_updates, extra_conflicts = self.merge_to(
+                    master.tags, overwrite=overwrite,
                                            source_refs=source_refs,
                                            ignore_master=ignore_master)
-            return conflicts
+                updates.update(extra_updates)
+                conflicts += extra_conflicts
+            return updates, conflicts
 
     def get_tag_dict(self):
         ret = {}
-        refs = self.get_refs()
+        refs = self.get_refs_container()
         for (name, peeled, unpeeled, bzr_revid) in self._iter_tag_refs(refs):
             ret[name] = bzr_revid
         return ret
@@ -168,38 +223,13 @@ class LocalGitTagDict(GitTags):
 
     def __init__(self, branch):
         super(LocalGitTagDict, self).__init__(branch)
-        self.refs = self.repository._git.refs
+        self.refs = self.repository.bzrdir._git.refs
 
-    def get_refs(self):
-        return self.repository._git.get_refs()
-
-    def _iter_tag_refs(self, refs):
-        """Iterate over the tag refs.
-
-        :param refs: Refs dictionary (name -> git sha1)
-        :return: iterator over (name, peeled_sha1, unpeeled_sha1, bzr_revid)
-        """
-        for k, (peeled, unpeeled) in extract_tags(refs).iteritems():
-            try:
-                obj = self.repository._git[peeled]
-            except KeyError:
-                mutter("Tag %s points at unknown object %s, ignoring", peeled,
-                       obj)
-                continue
-            # FIXME: this shouldn't really be necessary, the repository
-            # already should have these unpeeled.
-            while isinstance(obj, Tag):
-                peeled = obj.object[1]
-                obj = self.repository._git[peeled]
-            if not isinstance(obj, Commit):
-                mutter("Tag %s points at object %r that is not a commit, "
-                       "ignoring", k, obj)
-                continue
-            yield (k, peeled, unpeeled,
-                   self.branch.lookup_foreign_revision_id(peeled))
+    def get_refs_container(self):
+        return self.refs
 
     def _set_tag_dict(self, to_dict):
-        extra = set(self.get_refs().keys())
+        extra = set(self.refs.allkeys())
         for k, revid in to_dict.iteritems():
             name = tag_name_to_ref(k)
             if name in extra:
@@ -210,8 +240,11 @@ class LocalGitTagDict(GitTags):
                 del self.repository._git[name]
 
     def set_tag(self, name, revid):
-        self.refs[tag_name_to_ref(name)], _ = \
-            self.branch.lookup_bzr_revision_id(revid)
+        try:
+            git_sha, mapping = self.branch.lookup_bzr_revision_id(revid)
+        except errors.NoSuchRevision:
+            raise errors.GhostTagsNotSupported(self)
+        self.refs[tag_name_to_ref(name)] = git_sha
 
 
 class DictTagDict(tag.BasicTags):
@@ -224,6 +257,21 @@ class DictTagDict(tag.BasicTags):
         return self._tags
 
 
+class GitSymrefBranchFormat(branch.BranchFormat):
+
+    def get_format_description(self):
+        return 'Git Symbolic Reference Branch'
+
+    def network_name(self):
+        return "git"
+
+    def get_reference(self, controldir, name=None):
+        return controldir.get_branch_reference(name)
+
+    def set_reference(self, controldir, name, target):
+        return controldir.set_branch_reference(name, target)
+
+
 class GitBranchFormat(branch.BranchFormat):
 
     def get_format_description(self):
@@ -238,6 +286,12 @@ class GitBranchFormat(branch.BranchFormat):
     def supports_leaving_lock(self):
         return False
 
+    def supports_tags_referencing_ghosts(self):
+        return False
+
+    def tags_are_versioned(self):
+        return False
+
     @property
     def _matchingbzrdir(self):
         from bzrlib.plugins.git.dir import LocalGitControlDirFormat
@@ -248,21 +302,23 @@ class GitBranchFormat(branch.BranchFormat):
         return ForeignTestsBranchFactory()
 
     def make_tags(self, branch):
-        if getattr(branch.repository, "get_refs", None) is not None:
+        try:
+            return branch.tags
+        except AttributeError:
+            pass
+        if getattr(branch.repository, "_git", None) is None:
             from bzrlib.plugins.git.remote import RemoteGitTagDict
             return RemoteGitTagDict(branch)
         else:
             return LocalGitTagDict(branch)
 
-    def initialize(self, a_bzrdir, name=None, repository=None):
+    def initialize(self, a_bzrdir, name=None, repository=None,
+                   append_revisions_only=None):
         from bzrlib.plugins.git.dir import LocalGitDir
         if not isinstance(a_bzrdir, LocalGitDir):
             raise errors.IncompatibleFormat(self, a_bzrdir._format)
-        if repository is None:
-            repository = a_bzrdir.open_repository()
-        ref = branch_name_to_ref(name, "HEAD")
-        repository._git[ref] = ZERO_SHA
-        return LocalGitBranch(a_bzrdir, repository, ref, a_bzrdir._lockfiles)
+        return a_bzrdir.create_branch(repository=repository, name=name,
+            append_revisions_only=append_revisions_only)
 
 
 class GitReadLock(object):
@@ -285,22 +341,22 @@ class GitBranch(ForeignBranch):
     def control_transport(self):
         return self.bzrdir.control_transport
 
-    def __init__(self, bzrdir, repository, ref, lockfiles, tagsdict=None):
+    def __init__(self, bzrdir, repository, ref):
         self.base = bzrdir.root_transport.base
         self.repository = repository
         self._format = GitBranchFormat()
-        self.control_files = lockfiles
         self.bzrdir = bzrdir
         self._lock_mode = None
         self._lock_count = 0
         super(GitBranch, self).__init__(repository.get_mapping())
-        if tagsdict is not None:
-            self.tags = DictTagDict(self, tagsdict)
         self.ref = ref
-        self.name = ref_to_branch_name(ref)
+        try:
+            self.name = ref_to_branch_name(ref)
+        except ValueError:
+            self.name = None
         self._head = None
 
-    def _get_checkout_format(self):
+    def _get_checkout_format(self, lightweight=False):
         """Return the most suitable metadir for a checkout of this branch.
         Weaves are used if this branch's repository uses weaves.
         """
@@ -313,6 +369,12 @@ class GitBranch(ForeignBranch):
             return ret
         return "git"
 
+    def get_config(self):
+        return GitBranchConfig(self)
+
+    def get_config_stack(self):
+        return GitBranchStack(self)
+
     def _get_nick(self, local=False, possible_master_transports=None):
         """Find the nick name for this branch.
 
@@ -327,14 +389,14 @@ class GitBranch(ForeignBranch):
 
     def __repr__(self):
         return "<%s(%r, %r)>" % (self.__class__.__name__, self.repository.base,
-            self.ref or "HEAD")
+            self.name)
 
     def generate_revision_history(self, revid, old_revid=None):
         if revid == NULL_REVISION:
             newhead = ZERO_SHA
         else:
             # FIXME: Check that old_revid is in the ancestry of revid
-            newhead, self.mapping = self.mapping.revision_id_bzr_to_foreign(revid)
+            newhead, self.mapping = self.repository.lookup_bzr_revision_id(revid)
             if self.mapping is None:
                 raise AssertionError
         self._set_head(newhead)
@@ -343,7 +405,8 @@ class GitBranch(ForeignBranch):
         if token is not None:
             raise errors.TokenLockingNotSupported(self)
         if self._lock_mode:
-            assert self._lock_mode == 'w'
+            if self._lock_mode == 'r':
+                raise errors.ReadOnlyError(self)
             self._lock_count += 1
         else:
             self._lock_mode = 'w'
@@ -351,6 +414,12 @@ class GitBranch(ForeignBranch):
         self.repository.lock_write()
         return GitWriteLock(self.unlock)
 
+    def leave_lock_in_place(self):
+        raise NotImplementedError(self.leave_lock_in_place)
+
+    def dont_leave_lock_in_place(self):
+        raise NotImplementedError(self.dont_leave_lock_in_place)
+
     def get_stacked_on_url(self):
         # Git doesn't do stacking (yet...)
         raise errors.UnstackableBranchFormat(self._format, self.base)
@@ -364,6 +433,9 @@ class GitBranch(ForeignBranch):
         # FIXME: Set "origin" url in .git/config ?
         pass
 
+    def break_lock(self):
+        raise NotImplementedError(self.break_lock)
+
     def lock_read(self):
         if self._lock_mode:
             assert self._lock_mode in ('r', 'w')
@@ -414,11 +486,10 @@ class GitBranch(ForeignBranch):
 class LocalGitBranch(GitBranch):
     """A local Git branch."""
 
-    def __init__(self, bzrdir, repository, ref, lockfiles, tagsdict=None):
-        super(LocalGitBranch, self).__init__(bzrdir, repository, ref,
-              lockfiles, tagsdict)
-        refs = repository._git.get_refs()
-        if not (ref in refs.keys() or "HEAD" in refs.keys()):
+    def __init__(self, bzrdir, repository, ref):
+        super(LocalGitBranch, self).__init__(bzrdir, repository, ref)
+        refs = bzrdir.get_refs_container()
+        if not (ref in refs or "HEAD" in refs):
             raise errors.NotBranchError(self.base)
 
     def create_checkout(self, to_location, revision_id=None, lightweight=False,
@@ -426,7 +497,7 @@ class LocalGitBranch(GitBranch):
         if lightweight:
             t = transport.get_transport(to_location)
             t.ensure_base()
-            format = self._get_checkout_format()
+            format = self._get_checkout_format(lightweight=True)
             checkout = format.initialize_on_transport(t)
             from_branch = branch.BranchReferenceFormat().initialize(checkout,
                 self)
@@ -447,7 +518,8 @@ class LocalGitBranch(GitBranch):
         :return: WorkingTree object of checkout.
         """
         checkout_branch = bzrdir.BzrDir.create_branch_convenience(
-            to_location, force_new_tree=False)
+            to_location, force_new_tree=False,
+            format=self._get_checkout_format(lightweight=False))
         checkout = checkout_branch.bzrdir
         checkout_branch.bind(self)
         # pull up to the specified revision_id to set the initial
@@ -455,6 +527,10 @@ class LocalGitBranch(GitBranch):
         checkout_branch.pull(self, stop_revision=revision_id)
         return checkout.create_workingtree(revision_id, hardlink=hardlink)
 
+    def fetch(self, from_branch, last_revision=None, limit=None):
+        return branch.InterBranch.get(from_branch, self).fetch(
+            stop_revision=last_revision, limit=limit)
+
     def _gen_revision_history(self):
         if self.head is None:
             return []
@@ -499,9 +575,6 @@ class LocalGitBranch(GitBranch):
 
     head = property(_get_head, _set_head)
 
-    def get_config(self):
-        return GitBranchConfig(self)
-
     def get_push_location(self):
         """See Branch.get_push_location."""
         push_loc = self.get_config().get_user_option('push_location')
@@ -519,15 +592,24 @@ class LocalGitBranch(GitBranch):
 def _quick_lookup_revno(local_branch, remote_branch, revid):
     assert isinstance(revid, str), "was %r" % revid
     # Try in source branch first, it'll be faster
+    local_branch.lock_read()
     try:
-        return local_branch.revision_id_to_revno(revid)
-    except errors.NoSuchRevision:
-        graph = local_branch.repository.get_graph()
         try:
-            return graph.find_distance_to_null(revid)
-        except errors.GhostRevisionsHaveNoRevno:
-            # FIXME: Check using graph.find_distance_to_null() ?
-            return remote_branch.revision_id_to_revno(revid)
+            return local_branch.revision_id_to_revno(revid)
+        except errors.NoSuchRevision:
+            graph = local_branch.repository.get_graph()
+            try:
+                return graph.find_distance_to_null(revid,
+                    [(revision.NULL_REVISION, 0)])
+            except errors.GhostRevisionsHaveNoRevno:
+                # FIXME: Check using graph.find_distance_to_null() ?
+                remote_branch.lock_read()
+                try:
+                    return remote_branch.revision_id_to_revno(revid)
+                finally:
+                    remote_branch.unlock()
+    finally:
+        local_branch.unlock()
 
 
 class GitBranchPullResult(branch.PullResult):
@@ -623,14 +705,17 @@ class InterFromGitBranch(branch.GenericInterBranch):
             return False
         return True
 
-    def fetch(self, stop_revision=None, fetch_tags=True, limit=None):
+    def fetch(self, stop_revision=None, fetch_tags=None, limit=None):
         self.fetch_objects(stop_revision, fetch_tags=fetch_tags, limit=limit)
 
     def fetch_objects(self, stop_revision, fetch_tags, limit=None):
         interrepo = self._get_interrepo(self.source, self.target)
+        if fetch_tags is None:
+            c = self.source.get_config()
+            fetch_tags = c.get_user_option_as_bool('branch.fetch_tags')
         def determine_wants(heads):
             if self.source.ref is not None and not self.source.ref in heads:
-                raise NoSuchRef(self.source.ref, heads.keys())
+                raise NoSuchRef(self.source.ref, self.source.user_url, heads.keys())
 
             if stop_revision is None:
                 if self.source.ref is not None:
@@ -651,7 +736,7 @@ class InterFromGitBranch(branch.GenericInterBranch):
         return head, refs
 
     def _update_revisions(self, stop_revision=None, overwrite=False):
-        head, refs = self.fetch_objects(stop_revision, fetch_tags=True)
+        head, refs = self.fetch_objects(stop_revision, fetch_tags=None)
         if overwrite:
             prev_last_revid = None
         else:
@@ -660,6 +745,47 @@ class InterFromGitBranch(branch.GenericInterBranch):
             prev_last_revid, self.source)
         return head, refs
 
+    def _basic_pull(self, stop_revision, overwrite, run_hooks,
+              _override_hook_target, _hook_master):
+        result = GitBranchPullResult()
+        result.source_branch = self.source
+        if _override_hook_target is None:
+            result.target_branch = self.target
+        else:
+            result.target_branch = _override_hook_target
+        self.source.lock_read()
+        try:
+            self.target.lock_write()
+            try:
+                # We assume that during 'pull' the target repository is closer than
+                # the source one.
+                (result.old_revno, result.old_revid) = \
+                    self.target.last_revision_info()
+                result.new_git_head, remote_refs = self._update_revisions(
+                    stop_revision, overwrite=overwrite)
+                tags_ret  = self.source.tags.merge_to(
+                    self.target.tags, overwrite)
+                if isinstance(tags_ret, tuple):
+                    result.tag_updates, result.tag_conflicts = tags_ret
+                else:
+                    result.tag_conflicts = tags_ret
+                (result.new_revno, result.new_revid) = \
+                    self.target.last_revision_info()
+                if _hook_master:
+                    result.master_branch = _hook_master
+                    result.local_branch = result.target_branch
+                else:
+                    result.master_branch = result.target_branch
+                    result.local_branch = None
+                if run_hooks:
+                    for hook in branch.Branch.hooks['post_pull']:
+                        hook(result)
+                return result
+            finally:
+                self.target.unlock()
+        finally:
+            self.source.unlock()
+
     def pull(self, overwrite=False, stop_revision=None,
              possible_transports=None, _hook_master=None, run_hooks=True,
              _override_hook_target=None, local=False):
@@ -674,37 +800,38 @@ class InterFromGitBranch(branch.GenericInterBranch):
             supplied as the target_branch to pull hooks.
         """
         # This type of branch can't be bound.
-        if local:
+        bound_location = self.target.get_bound_location()
+        if local and not bound_location:
             raise errors.LocalRequiresBoundBranch()
-        result = GitBranchPullResult()
-        result.source_branch = self.source
-        if _override_hook_target is None:
-            result.target_branch = self.target
-        else:
-            result.target_branch = _override_hook_target
+        master_branch = None
+        source_is_master = False
         self.source.lock_read()
+        if bound_location:
+            # bound_location comes from a config file, some care has to be
+            # taken to relate it to source.user_url
+            normalized = urlutils.normalize_url(bound_location)
+            try:
+                relpath = self.source.user_transport.relpath(normalized)
+                source_is_master = (relpath == '')
+            except (errors.PathNotChild, errors.InvalidURL):
+                source_is_master = False
+        if not local and bound_location and not source_is_master:
+            # not pulling from master, so we need to update master.
+            master_branch = self.target.get_master_branch(possible_transports)
+            master_branch.lock_write()
         try:
-            # We assume that during 'pull' the target repository is closer than
-            # the source one.
-            (result.old_revno, result.old_revid) = \
-                self.target.last_revision_info()
-            result.new_git_head, remote_refs = self._update_revisions(
-                stop_revision, overwrite=overwrite)
-            result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
-                overwrite)
-            (result.new_revno, result.new_revid) = \
-                self.target.last_revision_info()
-            if _hook_master:
-                result.master_branch = _hook_master
-                result.local_branch = result.target_branch
-            else:
-                result.master_branch = result.target_branch
-                result.local_branch = None
-            if run_hooks:
-                for hook in branch.Branch.hooks['post_pull']:
-                    hook(result)
+            try:
+                if master_branch:
+                    # pull from source into master.
+                    master_branch.pull(self.source, overwrite, stop_revision,
+                        run_hooks=False)
+                result = self._basic_pull(stop_revision, overwrite, run_hooks,
+                    _override_hook_target, _hook_master=master_branch)
+            finally:
+                self.source.unlock()
         finally:
-            self.source.unlock()
+            if master_branch:
+                master_branch.unlock()
         return result
 
     def _basic_push(self, overwrite=False, stop_revision=None):
@@ -714,8 +841,12 @@ class InterFromGitBranch(branch.GenericInterBranch):
         result.old_revno, result.old_revid = self.target.last_revision_info()
         result.new_git_head, remote_refs = self._update_revisions(
             stop_revision, overwrite=overwrite)
-        result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
+        tags_ret = self.source.tags.merge_to(self.target.tags,
             overwrite)
+        if isinstance(tags_ret, tuple):
+            (result.tag_updates, result.tag_conflicts) = tags_ret
+        else:
+            result.tag_conflicts = tags_ret
         result.new_revno, result.new_revid = self.target.last_revision_info()
         return result
 
@@ -723,6 +854,9 @@ class InterFromGitBranch(branch.GenericInterBranch):
 class InterGitBranch(branch.GenericInterBranch):
     """InterBranch implementation that pulls between Git branches."""
 
+    def fetch(self, stop_revision=None, fetch_tags=None, limit=None):
+        raise NotImplementedError(self.fetch)
+
 
 class InterLocalGitRemoteGitBranch(InterGitBranch):
     """InterBranch that copies from a local to a remote git branch."""
@@ -772,14 +906,18 @@ class InterGitLocalGitBranch(InterGitBranch):
                 isinstance(target, LocalGitBranch))
 
     def _basic_push(self, overwrite=False, stop_revision=None):
-        result = branch.BranchPushResult()
+        result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
         result.old_revid = self.target.last_revision()
         refs, stop_revision = self.update_refs(stop_revision)
         self.target.generate_revision_history(stop_revision, result.old_revid)
-        result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
+        tags_ret = self.source.tags.merge_to(self.target.tags,
             source_refs=refs, overwrite=overwrite)
+        if isinstance(tags_ret, tuple):
+            (result.tag_updates, result.tag_conflicts) = tags_ret
+        else:
+            result.tag_conflicts = tags_ret
         result.new_revid = self.target.last_revision()
         return result
 
@@ -801,12 +939,29 @@ class InterGitLocalGitBranch(InterGitBranch):
         result = GitPullResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        result.old_revid = self.target.last_revision()
-        refs, stop_revision = self.update_refs(stop_revision)
-        self.target.generate_revision_history(stop_revision, result.old_revid)
-        result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
-            overwrite=overwrite, source_refs=refs)
-        result.new_revid = self.target.last_revision()
+        self.source.lock_read()
+        try:
+            self.target.lock_write()
+            try:
+                result.old_revid = self.target.last_revision()
+                refs, stop_revision = self.update_refs(stop_revision)
+                self.target.generate_revision_history(stop_revision, result.old_revid)
+                tags_ret = self.source.tags.merge_to(self.target.tags,
+                    overwrite=overwrite, source_refs=refs)
+                if isinstance(tags_ret, tuple):
+                    (result.tag_updates, result.tag_conflicts) = tags_ret
+                else:
+                    result.tag_conflicts = tags_ret
+                result.new_revid = self.target.last_revision()
+                result.local_branch = None
+                result.master_branch = result.target_branch
+                if run_hooks:
+                    for hook in branch.Branch.hooks['post_pull']:
+                        hook(result)
+            finally:
+                self.target.unlock()
+        finally:
+            self.source.unlock()
         return result
 
 
@@ -831,7 +986,8 @@ class InterToGitBranch(branch.GenericInterBranch):
         return (not isinstance(source, GitBranch) and
                 isinstance(target, GitBranch))
 
-    def _get_new_refs(self, stop_revision=None):
+    def _get_new_refs(self, stop_revision=None, fetch_tags=None):
+        assert self.source.is_locked()
         if stop_revision is None:
             (stop_revno, stop_revision) = self.source.last_revision_info()
         else:
@@ -839,30 +995,117 @@ class InterToGitBranch(branch.GenericInterBranch):
         assert type(stop_revision) is str
         main_ref = self.target.ref or "refs/heads/master"
         refs = { main_ref: (None, stop_revision) }
+        if fetch_tags is None:
+            c = self.source.get_config()
+            fetch_tags = c.get_user_option_as_bool('branch.fetch_tags')
         for name, revid in self.source.tags.get_tag_dict().iteritems():
             if self.source.repository.has_revision(revid):
-                refs[tag_name_to_ref(name)] = (None, revid)
+                ref = tag_name_to_ref(name)
+                if not check_ref_format(ref):
+                    warning("skipping tag with invalid characters %s (%s)",
+                        name, ref)
+                    continue
+                if fetch_tags:
+                    # FIXME: Skip tags that are not in the ancestry
+                    refs[ref] = (None, revid)
         return refs, main_ref, (stop_revno, stop_revision)
 
+    def _update_refs(self, result, old_refs, new_refs, overwrite):
+        mutter("updating refs. old refs: %r, new refs: %r",
+               old_refs, new_refs)
+        result.tag_updates = {}
+        result.tag_conflicts = []
+        ret = dict(old_refs)
+        def ref_equals(refs, ref, git_sha, revid):
+            try:
+                value = refs[ref]
+            except KeyError:
+                return False
+            if (value[0] is not None and
+                git_sha is not None and
+                value[0] == git_sha):
+                return True
+            if (value[1] is not None and
+                revid is not None and
+                value[1] == revid):
+                return True
+            # FIXME: If one side only has the git sha available and the other only
+            # has the bzr revid, then this will cause us to show a tag as updated
+            # that hasn't actually been updated. 
+            return False
+        # FIXME: Check for diverged branches
+        for ref, (git_sha, revid) in new_refs.iteritems():
+            if ref_equals(ret, ref, git_sha, revid):
+                # Already up to date
+                if git_sha is None:
+                    git_sha = old_refs[ref][0]
+                if revid is None:
+                    revid = old_refs[ref][1]
+                ret[ref] = new_refs[ref] = (git_sha, revid)
+            elif ref not in ret or overwrite:
+                try:
+                    tag_name = ref_to_tag_name(ref)
+                except ValueError:
+                    pass
+                else:
+                    result.tag_updates[tag_name] = revid
+                ret[ref] = (git_sha, revid)
+            else:
+                # FIXME: Check diverged
+                diverged = False
+                if diverged:
+                    try:
+                        name = ref_to_tag_name(ref)
+                    except ValueError:
+                        pass
+                    else:
+                        result.tag_conflicts.append((name, revid, ret[name][1]))
+                else:
+                    ret[ref] = (git_sha, revid)
+        return ret
+
+    def fetch(self, stop_revision=None, fetch_tags=None, lossy=False, limit=None):
+        assert limit is None
+        if stop_revision is None:
+            stop_revision = self.source.last_revision()
+        ret = []
+        if fetch_tags:
+            for k, v in self.source.tags.get_tag_dict().iteritems():
+                ret.append((None, v))
+        ret.append((None, stop_revision))
+        self.interrepo.fetch_objects(ret, roundtrip=(not lossy))
+
     def pull(self, overwrite=False, stop_revision=None, local=False,
              possible_transports=None, run_hooks=True):
         result = GitBranchPullResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
-        def update_refs(old_refs):
-            refs = dict(old_refs)
-            # FIXME: Check for diverged branches
-            refs.update(new_refs)
-            return refs
+        self.source.lock_read()
         try:
-            old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
-        except NoPushSupport:
-            raise errors.NoRoundtrippingSupport(self.source, self.target)
-        (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
-        if result.old_revid is None:
-            result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
-        result.new_revid = new_refs[main_ref][1]
+            self.target.lock_write()
+            try:
+                new_refs, main_ref, stop_revinfo = self._get_new_refs(
+                    stop_revision)
+                def update_refs(old_refs):
+                    return self._update_refs(result, old_refs, new_refs, overwrite)
+                try:
+                    result.revidmap, old_refs, new_refs = self.interrepo.fetch_refs(
+                        update_refs, lossy=False)
+                except NoPushSupport:
+                    raise errors.NoRoundtrippingSupport(self.source, self.target)
+                (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
+                if result.old_revid is None:
+                    result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
+                result.new_revid = new_refs[main_ref][1]
+                result.local_branch = None
+                result.master_branch = self.target
+                if run_hooks:
+                    for hook in branch.Branch.hooks['post_pull']:
+                        hook(result)
+            finally:
+                self.target.unlock()
+        finally:
+            self.source.unlock()
         return result
 
     def push(self, overwrite=False, stop_revision=None, lossy=False,
@@ -870,25 +1113,27 @@ class InterToGitBranch(branch.GenericInterBranch):
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
-        def update_refs(old_refs):
-            refs = dict(old_refs)
-            # FIXME: Check for diverged branches
-            refs.update(new_refs)
-            return refs
-        if lossy:
-            result.revidmap, old_refs, new_refs = self.interrepo.dfetch_refs(
-                update_refs)
-        else:
+        result.local_branch = None
+        result.master_branch = result.target_branch
+        self.source.lock_read()
+        try:
+            new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
+            def update_refs(old_refs):
+                return self._update_refs(result, old_refs, new_refs, overwrite)
             try:
-                old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+                result.revidmap, old_refs, new_refs = self.interrepo.fetch_refs(
+                    update_refs, lossy=lossy)
             except NoPushSupport:
                 raise errors.NoRoundtrippingSupport(self.source, self.target)
-        (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
-        if result.old_revid is None:
-            result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
-        result.new_revid = new_refs[main_ref][1]
-        (result.new_original_revno, result.new_original_revid) = stop_revinfo
+            (old_sha1, result.old_revid) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
+            if result.old_revid is None:
+                result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
+            result.new_revid = new_refs[main_ref][1]
+            (result.new_original_revno, result.new_original_revid) = stop_revinfo
+            for hook in branch.Branch.hooks['post_push']:
+                hook(result)
+        finally:
+            self.source.unlock()
         return result
 
     def lossy_push(self, stop_revision=None):
