@@ -18,12 +18,14 @@ from dulwich.objects import (
     Commit,
     Tag,
     Tree,
+    S_IFGITLINK,
     S_ISGITLINK,
     ZERO_SHA,
     )
 from dulwich.object_store import (
     tree_lookup_path,
     )
+from dulwich.walk import Walker
 from itertools import (
     imap,
     )
@@ -33,6 +35,7 @@ import stat
 
 from bzrlib import (
     debug,
+    errors,
     osutils,
     trace,
     ui,
@@ -68,6 +71,9 @@ from bzrlib.versionedfile import (
     ChunkedContentFactory,
     )
 
+from bzrlib.plugins.git.errors import (
+    NotCommitError,
+    )
 from bzrlib.plugins.git.mapping import (
     DEFAULT_FILE_MODE,
     mode_is_executable,
@@ -79,7 +85,9 @@ from bzrlib.plugins.git.object_store import (
     LRUTreeCache,
     _tree_to_objects,
     )
-from bzrlib.plugins.git.refs import extract_tags
+from bzrlib.plugins.git.refs import (
+    is_tag,
+    )
 from bzrlib.plugins.git.remote import (
     RemoteGitRepository,
     )
@@ -128,7 +136,7 @@ def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha),
         blob = lookup_object(hexsha)
         if ie.kind == "symlink":
             ie.revision = None
-            ie.symlink_target = blob.data
+            ie.symlink_target = blob.data.decode("utf-8")
         else:
             ie.text_size = sum(imap(len, blob.chunked))
             ie.text_sha1 = osutils.sha_strings(blob.chunked)
@@ -175,26 +183,33 @@ def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha),
 
 
 class SubmodulesRequireSubtrees(BzrError):
-    _fmt = """The repository you are fetching from contains submodules. To continue, upgrade your Bazaar repository to a format that supports nested trees, such as 'development-subtree'."""
+    _fmt = ("The repository you are fetching from contains submodules. "
+            "To continue, upgrade your Bazaar repository to a format that "
+            "supports nested trees, such as 'development-subtree'.")
     internal = False
 
 
 def import_git_submodule(texts, mapping, path, name, (base_hexsha, hexsha),
     base_inv, parent_id, revision_id, parent_invs, lookup_object,
     (base_mode, mode), store_updater, lookup_file_id):
+    """Import a git submodule."""
     if base_hexsha == hexsha and base_mode == mode:
         return [], {}
     file_id = lookup_file_id(path)
+    invdelta = []
     ie = TreeReference(file_id, name.decode("utf-8"), parent_id)
     ie.revision = revision_id
-    if base_hexsha is None:
-        oldpath = None
+    if base_hexsha is not None:
+        old_path = path.decode("utf-8") # Renames are not supported yet
+        if stat.S_ISDIR(base_mode):
+            invdelta.extend(remove_disappeared_children(base_inv, old_path,
+                lookup_object(base_hexsha), [], lookup_object))
     else:
-        oldpath = path
+        old_path = None
     ie.reference_revision = mapping.revision_id_foreign_to_bzr(hexsha)
     texts.insert_record_stream([
         ChunkedContentFactory((file_id, ie.revision), (), None, [])])
-    invdelta = [(oldpath, path, file_id, ie)]
+    invdelta.append((old_path, path, file_id, ie))
     return invdelta, {}
 
 
@@ -289,15 +304,19 @@ def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
                 file_id, revision_id, parent_invs, lookup_object,
                 (child_base_mode, child_mode), store_updater, lookup_file_id)
         else:
-            subinvdelta = import_git_blob(texts, mapping, child_path, name,
-                (child_base_hexsha, child_hexsha), base_inv, file_id,
-                revision_id, parent_invs, lookup_object,
-                (child_base_mode, child_mode), store_updater, lookup_file_id)
+            if not mapping.is_special_file(name):
+                subinvdelta = import_git_blob(texts, mapping, child_path, name,
+                    (child_base_hexsha, child_hexsha), base_inv, file_id,
+                    revision_id, parent_invs, lookup_object,
+                    (child_base_mode, child_mode), store_updater, lookup_file_id)
+            else:
+                subinvdelta = []
             grandchildmodes = {}
         child_modes.update(grandchildmodes)
         invdelta.extend(subinvdelta)
         if child_mode not in (stat.S_IFDIR, DEFAULT_FILE_MODE,
-                        stat.S_IFLNK, DEFAULT_FILE_MODE|0111):
+                        stat.S_IFLNK, DEFAULT_FILE_MODE|0111,
+                        S_IFGITLINK):
             child_modes[child_path] = child_mode
     # Remove any children that have disappeared
     if base_tree is not None and type(base_tree) is Tree:
@@ -322,7 +341,8 @@ def verify_commit_reconstruction(target_git_object_retriever, lookup_object,
     diff = []
     new_objs = {}
     for path, obj, ie in _tree_to_objects(ret_tree, parent_trees,
-        target_git_object_retriever._cache.idmap, unusual_modes, mapping.BZR_DUMMY_FILE):
+        target_git_object_retriever._cache.idmap, unusual_modes,
+        mapping.BZR_DUMMY_FILE):
         old_obj_id = tree_lookup_path(lookup_object, o.tree, path)[1]
         new_objs[path] = obj
         if obj.id != old_obj_id:
@@ -345,6 +365,14 @@ def verify_commit_reconstruction(target_git_object_retriever, lookup_object,
             old_obj, new_obj))
 
 
+def ensure_inventories_in_repo(repo, trees):
+    real_inv_vf = repo.inventories.without_fallbacks()
+    for t in trees:
+        revid = t.get_revision_id()
+        if not real_inv_vf.get_parent_map([(revid, )]):
+            repo.add_inventory(revid, t.inventory, t.get_parent_ids())
+
+
 def import_git_commit(repo, mapping, head, lookup_object,
                       target_git_object_retriever, trees_cache):
     o = lookup_object(head)
@@ -353,10 +381,14 @@ def import_git_commit(repo, mapping, head, lookup_object,
     # roundtrip data.
     rev, roundtrip_revid, verifiers = mapping.import_commit(
         o, mapping.revision_id_foreign_to_bzr)
+    if roundtrip_revid is not None:
+        original_revid = rev.revision_id
+        rev.revision_id = roundtrip_revid
     # We have to do this here, since we have to walk the tree and
     # we need to make sure to import the blobs / trees with the right
     # path; this may involve adding them more than once.
     parent_trees = trees_cache.revision_trees(rev.parent_ids)
+    ensure_inventories_in_repo(repo, parent_trees)
     if parent_trees == []:
         base_inv = Inventory(root_id=None)
         base_tree = None
@@ -366,13 +398,14 @@ def import_git_commit(repo, mapping, head, lookup_object,
         base_tree = lookup_object(o.parents[0]).tree
         base_mode = stat.S_IFDIR
     store_updater = target_git_object_retriever._get_updater(rev)
-    fileid_map = mapping.get_fileid_map(lookup_object, o.tree)
+    tree_supplement = mapping.get_fileid_map(lookup_object, o.tree)
     inv_delta, unusual_modes = import_git_tree(repo.texts,
             mapping, "", "", (base_tree, o.tree), base_inv,
             None, rev.revision_id, [p.inventory for p in parent_trees],
             lookup_object, (base_mode, stat.S_IFDIR), store_updater,
-            fileid_map.lookup_file_id,
-            allow_submodules=getattr(repo._format, "supports_tree_reference", False))
+            tree_supplement.lookup_file_id,
+            allow_submodules=getattr(repo._format, "supports_tree_reference",
+                False))
     if unusual_modes != {}:
         for path, mode in unusual_modes.iteritems():
             warn_unusual_mode(rev.foreign_revid, path, mode)
@@ -385,10 +418,8 @@ def import_git_commit(repo, mapping, head, lookup_object,
     rev.inventory_sha1, inv = repo.add_inventory_by_delta(basis_id,
               inv_delta, rev.revision_id, rev.parent_ids, base_inv)
     ret_tree = InventoryRevisionTree(repo, inv, rev.revision_id)
-    if roundtrip_revid is not None:
-        original_revid = rev.revision_id
-        rev.revision_id = roundtrip_revid
-        # Check verifiers
+    # Check verifiers
+    if verifiers and roundtrip_revid is not None:
         if getattr(StrictTestament3, "from_revision_tree", None):
             testament = StrictTestament3(rev, ret_tree)
         else: # bzr < 2.4
@@ -399,6 +430,9 @@ def import_git_commit(repo, mapping, head, lookup_object,
                          calculated_verifiers["testament3-sha1"],
                          rev.revision_id, verifiers["testament3-sha1"])
             rev.revision_id = original_revid
+            rev.inventory_sha1, inv = repo.add_inventory_by_delta(basis_id,
+              inv_delta, rev.revision_id, rev.parent_ids, base_inv)
+            ret_tree = InventoryRevisionTree(repo, inv, rev.revision_id)
     else:
         calculated_verifiers = {}
     store_updater.add_object(o, calculated_verifiers, None)
@@ -434,7 +468,9 @@ def import_git_objects(repo, mapping, object_iter,
         if pb is not None:
             pb.update("finding revisions to fetch", len(graph), None)
         head = heads.pop()
-        assert isinstance(head, str)
+        if head == ZERO_SHA:
+            continue
+        assert isinstance(head, str), "head is %r" % (head,)
         try:
             o = lookup_object(head)
         except KeyError:
@@ -490,7 +526,7 @@ def import_git_objects(repo, mapping, object_iter,
     return pack_hints, last_imported
 
 
-class InterGitRepository(InterRepository):
+class InterFromGitRepository(InterRepository):
 
     _matching_repo_format = GitRepositoryFormat()
 
@@ -498,56 +534,96 @@ class InterGitRepository(InterRepository):
         raise NotImplementedError(self._target_has_shas)
 
     def get_determine_wants_heads(self, wants, include_tags=False):
-        wants = set(wants)
-        def determine_wants(refs):
-            potential = set(wants)
-            if include_tags:
-                potential.update(
-                    [v[1] or v[0] for v in extract_tags(refs).itervalues()])
-            return list(potential - self._target_has_shas(potential))
-        return determine_wants
+        raise NotImplementedError(self.get_determine_wants_heads)
 
     def determine_wants_all(self, refs):
-        potential = set([sha for (ref, sha) in refs.iteritems() if not
-            ref.endswith("^{}")])
-        return list(potential - self._target_has_shas(potential))
+        raise NotImplementedError(self.determine_wants_all)
 
     @staticmethod
     def _get_repo_format_to_test():
         return None
 
-    def copy_content(self, revision_id=None, pb=None):
+    def copy_content(self, revision_id=None):
         """See InterRepository.copy_content."""
-        self.fetch(revision_id, pb, find_ghosts=False)
+        self.fetch(revision_id, find_ghosts=False)
+
+    def search_missing_revision_ids(self,
+            find_ghosts=True, revision_ids=None, if_present_ids=None,
+            limit=None):
+        git_shas = []
+        todo = []
+        if revision_ids:
+            todo.extend(revision_ids)
+        if if_present_ids:
+            todo.extend(revision_ids)
+        for revid in revision_ids:
+            if revid == NULL_REVISION:
+                continue
+            git_sha, mapping = self.source.lookup_bzr_revision_id(revid)
+            git_shas.append(git_sha)
+        walker = Walker(self.source._git.object_store,
+            include=git_shas, exclude=[sha for sha in self.target.bzrdir.get_refs_container().as_dict().values() if sha != ZERO_SHA])
+        missing_revids = set()
+        for entry in walker:
+            missing_revids.add(self.source.lookup_foreign_revision_id(entry.commit.id))
+        return self.source.revision_ids_to_search_result(missing_revids)
 
 
-class InterGitNonGitRepository(InterGitRepository):
+class InterGitNonGitRepository(InterFromGitRepository):
     """Base InterRepository that copies revisions from a Git into a non-Git
     repository."""
 
     def _target_has_shas(self, shas):
-        revids = [self.source.lookup_foreign_revision_id(sha) for sha in shas]
-        return self.target.has_revisions(revids)
+        revids = {}
+        for sha in shas:
+            try:
+                revid = self.source.lookup_foreign_revision_id(sha)
+            except NotCommitError:
+                # Commit is definitely not present
+                continue
+            else:
+                revids[revid] = sha
+        return set([revids[r] for r in self.target.has_revisions(revids)])
+
+    def determine_wants_all(self, refs):
+        potential = set()
+        for k, v in refs.as_dict().iteritems():
+            # For non-git target repositories, only worry about peeled
+            potential.add(self.source.bzrdir.get_peeled(k))
+        return list(potential - self._target_has_shas(potential))
+
+    def get_determine_wants_heads(self, wants, include_tags=False):
+        wants = set(wants)
+        def determine_wants(refs):
+            potential = set(wants)
+            if include_tags:
+                for k, unpeeled in refs.as_dict().iteritems():
+                    if not is_tag(k):
+                        continue
+                    potential.add(self.source.bzrdir.get_peeled(k))
+            return list(potential - self._target_has_shas(potential))
+        return determine_wants
 
     def get_determine_wants_revids(self, revids, include_tags=False):
         wants = set()
         for revid in set(revids):
+            if self.target.has_revision(revid):
+                continue
             git_sha, mapping = self.source.lookup_bzr_revision_id(revid)
             wants.add(git_sha)
         return self.get_determine_wants_heads(wants, include_tags=include_tags)
 
-    def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
+    def fetch_objects(self, determine_wants, mapping, limit=None):
         """Fetch objects from a remote server.
 
         :param determine_wants: determine_wants callback
         :param mapping: BzrGitMapping to use
-        :param pb: Optional progress bar
         :param limit: Maximum number of commits to import.
         :return: Tuple with pack hint, last imported revision id and remote refs
         """
         raise NotImplementedError(self.fetch_objects)
 
-    def fetch(self, revision_id=None, pb=None, find_ghosts=False,
+    def fetch(self, revision_id=None, find_ghosts=False,
               mapping=None, fetch_spec=None):
         if mapping is None:
             mapping = self.source.get_mapping()
@@ -564,16 +640,15 @@ class InterGitNonGitRepository(InterGitRepository):
             interesting_heads = None
 
         if interesting_heads is not None:
-            determine_wants = self.get_determine_wants_revids(interesting_heads,
-                include_tags=False)
+            determine_wants = self.get_determine_wants_revids(
+                interesting_heads, include_tags=False)
         else:
             determine_wants = self.determine_wants_all
 
         (pack_hint, _, remote_refs) = self.fetch_objects(determine_wants,
-            mapping, pb)
+            mapping)
         if pack_hint is not None and self.target._format.pack_compresses:
             self.target.pack(hint=pack_hint)
-        assert isinstance(remote_refs, dict)
         return remote_refs
 
 
@@ -613,10 +688,8 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
         map(all_parents.update, parent_map.itervalues())
         return set(all_revs) - all_parents
 
-    def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
+    def fetch_objects(self, determine_wants, mapping, limit=None):
         """See `InterGitNonGitRepository`."""
-        def progress(text):
-            report_git_progress(pb, text)
         store = BazaarObjectStore(self.target, mapping)
         store.lock_write()
         try:
@@ -625,13 +698,11 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
                     [store._lookup_revision_sha1(head) for head in heads])
             wants_recorder = DetermineWantsRecorder(determine_wants)
 
-            create_pb = None
-            if pb is None:
-                create_pb = pb = ui.ui_factory.nested_progress_bar()
+            pb = ui.ui_factory.nested_progress_bar()
             try:
                 objects_iter = self.source.fetch_objects(
                     wants_recorder, graph_walker, store.get_raw,
-                    progress)
+                    progress=lambda text: report_git_progress(pb, text))
                 trace.mutter("Importing %d new revisions",
                              len(wants_recorder.wants))
                 (pack_hint, last_rev) = import_git_objects(self.target,
@@ -639,8 +710,7 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
                     limit)
                 return (pack_hint, last_rev, wants_recorder.remote_refs)
             finally:
-                if create_pb:
-                    create_pb.finished()
+                pb.finished()
         finally:
             store.unlock()
 
@@ -662,13 +732,12 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
     """InterRepository that copies revisions from a local Git into a non-Git
     repository."""
 
-    def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
+    def fetch_objects(self, determine_wants, mapping, limit=None):
         """See `InterGitNonGitRepository`."""
-        remote_refs = self.source._git.get_refs()
+        remote_refs = self.source.bzrdir.get_refs_container()
         wants = determine_wants(remote_refs)
         create_pb = None
-        if pb is None:
-            create_pb = pb = ui.ui_factory.nested_progress_bar()
+        pb = ui.ui_factory.nested_progress_bar()
         target_git_object_retriever = BazaarObjectStore(self.target, mapping)
         try:
             target_git_object_retriever.lock_write()
@@ -680,8 +749,7 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
             finally:
                 target_git_object_retriever.unlock()
         finally:
-            if create_pb:
-                create_pb.finished()
+            pb.finished()
 
     @staticmethod
     def is_compatible(source, target):
@@ -697,52 +765,79 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
         return True
 
 
-class InterGitGitRepository(InterGitRepository):
+class InterGitGitRepository(InterFromGitRepository):
     """InterRepository that copies between Git repositories."""
 
-    def fetch_objects(self, determine_wants, mapping, pb=None):
-        def progress(text):
-            trace.note("git: %s", text)
+    def fetch_refs(self, update_refs, lossy=False):
+        if lossy:
+            raise errors.LossyPushToSameVCS(self.source, self.target)
+        old_refs = self.target.bzrdir.get_refs_container()
+        ref_changes = {}
+        def determine_wants(heads):
+            old_refs = dict([(k, (v, None)) for (k, v) in heads.iteritems()])
+            new_refs = update_refs(old_refs)
+            ref_changes.update(new_refs)
+            return [sha1 for (sha1, bzr_revid) in new_refs.itervalues()]
+        self.fetch_objects(determine_wants)
+        for k, (git_sha, bzr_revid) in ref_changes.iteritems():
+            self.target._git.refs[k] = git_sha
+        new_refs = self.target.bzrdir.get_refs_container()
+        return None, old_refs, new_refs
+
+    def fetch_objects(self, determine_wants, mapping=None):
         graphwalker = self.target._git.get_graph_walker()
         if (isinstance(self.source, LocalGitRepository) and
             isinstance(self.target, LocalGitRepository)):
-            refs = self.source._git.fetch(self.target._git, determine_wants,
-                progress)
+            def wrap_determine_wants(refs):
+                return determine_wants(self.source._git.refs)
+            pb = ui.ui_factory.nested_progress_bar()
+            try:
+                refs = self.source._git.fetch(self.target._git, wrap_determine_wants,
+                    lambda text: report_git_progress(pb, text))
+            finally:
+                pb.finished()
             return (None, None, refs)
         elif (isinstance(self.source, LocalGitRepository) and
               isinstance(self.target, RemoteGitRepository)):
             raise NotImplementedError
         elif (isinstance(self.source, RemoteGitRepository) and
               isinstance(self.target, LocalGitRepository)):
-            f, commit = self.target._git.object_store.add_thin_pack()
+            pb = ui.ui_factory.nested_progress_bar()
             try:
-                refs = self.source.bzrdir.root_transport.fetch_pack(
-                    determine_wants, graphwalker, f.write, progress)
-                commit()
-                return (None, None, refs)
-            except:
-                f.close()
-                raise
+                f, commit = self.target._git.object_store.add_pack()
+                try:
+                    refs = self.source.bzrdir.fetch_pack(
+                        determine_wants, graphwalker, f.write,
+                        lambda text: report_git_progress(pb, text))
+                    commit()
+                    return (None, None, refs)
+                except:
+                    f.close()
+                    raise
+            finally:
+                pb.finished()
         else:
-            raise AssertionError
+            raise AssertionError("fetching between %r and %r not supported" %
+                    (self.source, self.target))
 
     def _target_has_shas(self, shas):
-        return set([sha for sha in shas if self.target._git.object_store])
+        return set([sha for sha in shas if sha in self.target._git.object_store])
 
-    def fetch(self, revision_id=None, pb=None, find_ghosts=False,
+    def fetch(self, revision_id=None, find_ghosts=False,
               mapping=None, fetch_spec=None, branches=None):
         if mapping is None:
             mapping = self.source.get_mapping()
         r = self.target._git
         if revision_id is not None:
-            args = [mapping.revision_id_bzr_to_foreign(revision_id)[0]]
+            args = [self.source.lookup_bzr_revision_id(revision_id)[0]]
         elif fetch_spec is not None:
             recipe = fetch_spec.get_recipe()
             if recipe[0] in ("search", "proxy-search"):
                 heads = recipe[1]
             else:
-                raise AssertionError("Unsupported search result type %s" % recipe[0])
-            args = [mapping.revision_id_bzr_to_foreign(revid)[0] for revid in heads]
+                raise AssertionError(
+                    "Unsupported search result type %s" % recipe[0])
+            args = [self.source.lookup_bzr_revision_id(revid)[0] for revid in heads]
         if branches is not None:
             determine_wants = lambda x: [x[y] for y in branches if not x[y] in r.object_store and x[y] != ZERO_SHA]
         elif fetch_spec is None and revision_id is None:
@@ -762,6 +857,27 @@ class InterGitGitRepository(InterGitRepository):
     def get_determine_wants_revids(self, revids, include_tags=False):
         wants = set()
         for revid in set(revids):
+            if self.target.has_revision(revid):
+                continue
             git_sha, mapping = self.source.lookup_bzr_revision_id(revid)
             wants.add(git_sha)
-        return self.get_determine_wants_heads(wants, include_tags=include_tags)
+        return self.get_determine_wants_heads(wants,
+            include_tags=include_tags)
+
+    def determine_wants_all(self, refs):
+        potential = set(refs.as_dict().values())
+        return list(potential - self._target_has_shas(potential))
+
+    def get_determine_wants_heads(self, wants, include_tags=False):
+        wants = set(wants)
+        def determine_wants(refs):
+            potential = set(wants)
+            if include_tags:
+                for k, unpeeled in refs.as_dict().iteritems():
+                    if not is_tag(k):
+                        continue
+                    potential.add(unpeeled)
+            return list(potential - self._target_has_shas(potential))
+        return determine_wants
+
+
