@@ -50,6 +50,7 @@ from bzrlib.plugins.git.mapping import (
     directory_to_tree,
     extract_unusual_modes,
     mapping_registry,
+    object_mode,
     symlink_to_blob,
     )
 from bzrlib.plugins.git.unpeel_map import (
@@ -89,7 +90,6 @@ class LRUTreeCache(object):
         except KeyError:
             tree = self.repository.revision_tree(revid)
             self.add(tree)
-        assert tree.get_revision_id() == tree.inventory.revision_id
         return tree
 
     def iter_revision_trees(self, revids):
@@ -112,7 +112,7 @@ class LRUTreeCache(object):
         return list(self.iter_revision_trees(revids))
 
     def add(self, tree):
-        self._cache.add(tree.get_revision_id(), tree)
+        self._cache[tree.get_revision_id()] = tree
 
 
 def _find_missing_bzr_revids(graph, want, have):
@@ -225,7 +225,7 @@ def _tree_to_objects(tree, parent_trees, idmap, unusual_modes,
         elif kind[1] not in (None, "directory"):
             raise AssertionError(kind[1])
         if (path[0] not in (None, "") and
-            parent[0] in tree.inventory and
+            tree.has_id(parent[0]) and
             tree.inventory[parent[0]].kind == "directory"):
             # Removal
             new_trees[posixpath.dirname(path[0])] = parent[0]
@@ -291,8 +291,12 @@ class PackTupleIterable(object):
 
     def __init__(self, store):
         self.store = store
+        self.store.lock_read()
         self.objects = {}
 
+    def __del__(self):
+        self.store.unlock()
+
     def add(self, sha, path):
         self.objects[sha] = path
 
@@ -323,27 +327,30 @@ class BazaarObjectStore(BaseObjectStore):
         self.tree_cache = LRUTreeCache(self.repository)
         self.unpeel_map = UnpeelMap.from_repository(self.repository)
 
+    def _missing_revisions(self, revisions):
+        return self._cache.idmap.missing_revisions(revisions)
+
     def _update_sha_map(self, stop_revision=None):
         if not self.is_locked():
             raise AssertionError()
         if self._map_updated:
             return
         if (stop_revision is not None and
-            not self._cache.idmap.missing_revisions([stop_revision])):
+            not self._missing_revisions([stop_revision])):
             return
         graph = self.repository.get_graph()
         if stop_revision is None:
             all_revids = self.repository.all_revision_ids()
-            missing_revids = self._cache.idmap.missing_revisions(all_revids)
+            missing_revids = self._missing_revisions(all_revids)
         else:
             heads = set([stop_revision])
-            missing_revids = self._cache.idmap.missing_revisions(heads)
+            missing_revids = self._missing_revisions(heads)
             while heads:
                 parents = graph.get_parent_map(heads)
                 todo = set()
                 for p in parents.values():
                     todo.update([x for x in p if x not in missing_revids])
-                heads = self._cache.idmap.missing_revisions(todo)
+                heads = self._missing_revisions(todo)
                 missing_revids.update(heads)
         if NULL_REVISION in missing_revids:
             missing_revids.remove(NULL_REVISION)
@@ -391,11 +398,11 @@ class BazaarObjectStore(BaseObjectStore):
         return self.mapping.export_commit(rev, tree_sha, parent_lookup,
             roundtrip, verifiers)
 
-    def _create_fileid_map_blob(self, inv):
+    def _create_fileid_map_blob(self, tree):
         # FIXME: This can probably be a lot more efficient, 
         # not all files necessarily have to be processed.
         file_ids = {}
-        for (path, ie) in inv.iter_entries():
+        for (path, ie) in tree.inventory.iter_entries():
             if self.mapping.generate_file_id(path) != ie.file_id:
                 file_ids[path] = ie.file_id
         return self.mapping.export_fileid_map(file_ids)
@@ -429,7 +436,7 @@ class BazaarObjectStore(BaseObjectStore):
                 root_tree = self[self[base_sha1].tree]
             root_ie = tree.inventory.root
         if roundtrip and self.mapping.BZR_FILE_IDS_FILE is not None:
-            b = self._create_fileid_map_blob(tree.inventory)
+            b = self._create_fileid_map_blob(tree)
             if b is not None:
                 root_tree[self.mapping.BZR_FILE_IDS_FILE] = (
                     (stat.S_IFREG | 0644), b.id)
@@ -493,7 +500,7 @@ class BazaarObjectStore(BaseObjectStore):
             _check_expected_sha(expected_sha, blob)
             yield blob
 
-    def _reconstruct_tree(self, fileid, revid, inv, unusual_modes,
+    def _reconstruct_tree(self, fileid, revid, bzr_tree, unusual_modes,
         expected_sha=None):
         """Return a Git Tree object from a file id and a revision stored in bzr.
 
@@ -506,7 +513,7 @@ class BazaarObjectStore(BaseObjectStore):
                     return self._cache.idmap.lookup_tree_id(entry.file_id,
                         revid)
                 except (NotImplementedError, KeyError):
-                    obj = self._reconstruct_tree(entry.file_id, revid, inv,
+                    obj = self._reconstruct_tree(entry.file_id, revid, bzr_tree,
                         unusual_modes)
                     if obj is None:
                         return None
@@ -522,13 +529,13 @@ class BazaarObjectStore(BaseObjectStore):
                         [(entry.file_id, entry.revision, None)]).next().id
             else:
                 raise AssertionError("unknown entry kind '%s'" % entry.kind)
-        tree = directory_to_tree(inv[fileid], get_ie_sha1, unusual_modes,
+        tree = directory_to_tree(bzr_tree.inventory[fileid], get_ie_sha1, unusual_modes,
             self.mapping.BZR_DUMMY_FILE)
-        if (inv.root.file_id == fileid and
+        if (bzr_tree.get_root_id() == fileid and
             self.mapping.BZR_FILE_IDS_FILE is not None):
             if tree is None:
                 tree = Tree()
-            b = self._create_fileid_map_blob(inv)
+            b = self._create_fileid_map_blob(bzr_tree)
             # If this is the root tree, add the file ids
             tree[self.mapping.BZR_FILE_IDS_FILE] = (
                 (stat.S_IFREG | 0644), b.id)
@@ -639,6 +646,9 @@ class BazaarObjectStore(BaseObjectStore):
                 try:
                     rev = self.repository.get_revision(revid)
                 except errors.NoSuchRevision:
+                    if revid == NULL_REVISION:
+                        raise AssertionError(
+                            "should not try to look up NULL_REVISION")
                     trace.mutter('entry for %s %s in shamap: %r, but not '
                                  'found in repository', kind, sha, type_data)
                     raise KeyError(sha)
@@ -656,12 +666,13 @@ class BazaarObjectStore(BaseObjectStore):
                     tree = self.tree_cache.revision_tree(revid)
                     rev = self.repository.get_revision(revid)
                 except errors.NoSuchRevision:
-                    trace.mutter('entry for %s %s in shamap: %r, but not found in repository', kind, sha, type_data)
+                    trace.mutter('entry for %s %s in shamap: %r, but not found in '
+                        'repository', kind, sha, type_data)
                     raise KeyError(sha)
                 unusual_modes = extract_unusual_modes(rev)
                 try:
                     return self._reconstruct_tree(fileid, revid,
-                        tree.inventory, unusual_modes, expected_sha=sha)
+                        tree, unusual_modes, expected_sha=sha)
                 except errors.NoSuchRevision:
                     raise KeyError(sha)
             else:
@@ -704,7 +715,6 @@ class BazaarObjectStore(BaseObjectStore):
 
         graph = self.repository.get_graph()
         todo = _find_missing_bzr_revids(graph, pending, processed)
-        trace.mutter('sending revisions %r', todo)
         ret = PackTupleIterable(self)
         pb = ui.ui_factory.nested_progress_bar()
         try:
