@@ -16,6 +16,9 @@
 
 """Push implementation that simply prints message saying push is not supported."""
 
+from dulwich.objects import ZERO_SHA
+from dulwich.walk import Walker
+
 from bzrlib import (
     errors,
     ui,
@@ -86,6 +89,9 @@ class MissingObjectsIterator(object):
             if obj.type_name == "commit":
                 commit = obj
             self._pending.append((obj, path))
+        if commit is None:
+            raise AssertionError("no commit object generated for revision %s" %
+                revid)
         return commit.id
 
     def __len__(self):
@@ -113,34 +119,60 @@ class InterToGitRepository(InterRepository):
         """See InterRepository.copy_content."""
         self.fetch(revision_id, pb, find_ghosts=False)
 
-    def dfetch_refs(self, update_refs):
-        """Fetch non-roundtripped revisions into the target repository.
-
-        :param update_refs: Generate refs to fetch. Receives dictionary
-            with old names to old git shas. Should return a dictionary
-            of new names to Bazaar revision ids.
-        :return: revision id map, old refs dictionary and new refs dictionary
-        """
-        raise NotImplementedError(self.dfetch_refs)
-
-    def fetch_refs(self, update_refs):
-        """Fetch possibly roundtripped revisions into the target repository.
+    def fetch_refs(self, update_refs, lossy):
+        """Fetch possibly roundtripped revisions into the target repository
+        and update refs.
 
         :param update_refs: Generate refs to fetch. Receives dictionary
             with old refs (git shas), returns dictionary of new names to
             git shas.
+        :param lossy: Whether to roundtrip
         :return: old refs, new refs
         """
         raise NotImplementedError(self.fetch_refs)
 
+    def search_missing_revision_ids(self,
+            find_ghosts=True, revision_ids=None, if_present_ids=None,
+            limit=None):
+        git_shas = []
+        todo = []
+        if revision_ids:
+            todo.extend(revision_ids)
+        if if_present_ids:
+            todo.extend(revision_ids)
+        self.source_store.lock_read()
+        try:
+            for revid in revision_ids:
+                if revid == NULL_REVISION:
+                    continue
+                git_sha = self.source_store._lookup_revision_sha1(revid)
+                git_shas.append(git_sha)
+            walker = Walker(self.source_store,
+                include=git_shas, exclude=[sha for sha in self.target.bzrdir.get_refs_container().as_dict().values() if sha != ZERO_SHA])
+            missing_revids = set()
+            for entry in walker:
+                for (kind, type_data) in self.source_store.lookup_git_sha(entry.commit.id):
+                    if kind == "commit":
+                        missing_revids.add(type_data[0])
+        finally:
+            self.source_store.unlock()
+        return self.source.revision_ids_to_search_result(missing_revids)
+
 
 class InterToLocalGitRepository(InterToGitRepository):
     """InterBranch implementation between a Bazaar and a Git repository."""
 
     def __init__(self, source, target):
         super(InterToLocalGitRepository, self).__init__(source, target)
-        self.target_store = self.target._git.object_store
-        self.target_refs = self.target._git.refs
+        self.target_store = self.target.bzrdir._git.object_store
+        self.target_refs = self.target.bzrdir._git.refs
+
+    def _commit_needs_fetching(self, sha_id):
+        try:
+            return (sha_id not in self.target_store)
+        except errors.NoSuchRevision:
+            # Ghost, can't push
+            return False
 
     def _revision_needs_fetching(self, sha_id, revid):
         if revid == NULL_REVISION:
@@ -150,11 +182,7 @@ class InterToLocalGitRepository(InterToGitRepository):
                 sha_id = self.source_store._lookup_revision_sha1(revid)
             except KeyError:
                 return False
-        try:
-            return (sha_id not in self.target_store)
-        except errors.NoSuchRevision:
-            # Ghost, can't push
-            return False
+        return self._commit_needs_fetching(sha_id)
 
     def missing_revisions(self, stop_revisions):
         """Find the revisions that are missing from the target repository.
@@ -167,34 +195,37 @@ class InterToLocalGitRepository(InterToGitRepository):
         """
         revid_sha_map = {}
         stop_revids = []
-        stop_sha1s = []
         for (sha1, revid) in stop_revisions:
             if sha1 is not None and revid is not None:
                 revid_sha_map[revid] = sha1
+                stop_revids.append(revid)
             elif sha1 is not None:
-                stop_sha1s.append(sha1)
+                if self._commit_needs_fetching(sha1):
+                    for (kind, (revid, tree_sha, verifiers)) in self.source_store.lookup_git_sha(sha1):
+                        revid_sha_map[revid] = sha1
+                        stop_revids.append(revid)
             else:
                 assert revid is not None
                 stop_revids.append(revid)
-        missing = []
+        missing = set()
         graph = self.source.get_graph()
         pb = ui.ui_factory.nested_progress_bar()
         try:
-            for revid, _ in graph.iter_ancestry(stop_revids):
-                assert type(revid) is str
+            while stop_revids:
+                new_stop_revids = []
+                for revid in stop_revids:
+                    sha1 = revid_sha_map.get(revid)
+                    if (not revid in missing and
+                        self._revision_needs_fetching(sha1, revid)):
+                        missing.add(revid)
+                        new_stop_revids.append(revid)
+                stop_revids = set()
+                parent_map = graph.get_parent_map(new_stop_revids)
+                for parent_revids in parent_map.itervalues():
+                    stop_revids.update(parent_revids)
                 pb.update("determining revisions to fetch", len(missing))
-                sha1 = revid_sha_map.get(revid)
-                if self._revision_needs_fetching(sha1, revid):
-                    missing.append(revid)
         finally:
             pb.finished()
-        for sha1 in stop_sha1s:
-            try:
-                for (kind, (revid, tree_sha, verifiers)) in self.source_store.lookup_git_sha(sha1):
-                    missing.append(revid)
-                    revid_sha_map[revid] = sha1
-            except KeyError:
-                continue
         return graph.iter_topo_order(missing)
 
     def _get_target_bzr_refs(self):
@@ -219,84 +250,69 @@ class InterToLocalGitRepository(InterToGitRepository):
             bzr_refs[k] = (v, revid)
         return bzr_refs
 
-    def fetch_refs(self, update_refs):
+    def fetch_refs(self, update_refs, lossy):
         self.source_store.lock_read()
         try:
             old_refs = self._get_target_bzr_refs()
             new_refs = update_refs(old_refs)
-            self.fetch(mapped_refs=new_refs.values())
-        finally:
-            self.source_store.unlock()
-        return old_refs, new_refs
-
-    def dfetch_refs(self, update_refs):
-        self.source_store.lock_read()
-        try:
-            old_refs = self._get_target_bzr_refs()
-            new_refs = update_refs(old_refs)
-            gitidmap = {}
-            revidmap = {}
-            todo = list(self.missing_revisions(new_refs.values()))
-            pb = ui.ui_factory.nested_progress_bar()
-            try:
-                object_generator = self._get_missing_objects_iterator(pb)
-                for old_bzr_revid, git_commit in object_generator.import_revisions(
-                    todo, roundtrip=False):
-                    new_bzr_revid = self.mapping.revision_id_foreign_to_bzr(git_commit)
-                    assert type(old_bzr_revid) is str
-                    assert type(new_bzr_revid) is str
-                    assert type(git_commit) is str
-                    revidmap[old_bzr_revid] = new_bzr_revid
-                    gitidmap[old_bzr_revid] = git_commit
-                self.target_store.add_objects(object_generator)
-            finally:
-                pb.finished()
+            revidmap = self.fetch_objects(
+                [(git_sha, bzr_revid) for (git_sha, bzr_revid) in new_refs.values() if git_sha is None or not git_sha.startswith('ref:')], roundtrip=not lossy)
             for name, (gitid, revid) in new_refs.iteritems():
                 if gitid is None:
                     try:
-                        gitid = gitidmap[revid]
+                        gitid = revidmap[revid][0]
                     except KeyError:
                         gitid = self.source_store._lookup_revision_sha1(revid)
+                assert len(gitid) == 40 or gitid.startswith('ref: ')
                 self.target_refs[name] = gitid
         finally:
             self.source_store.unlock()
         return revidmap, old_refs, new_refs
 
-    def _get_missing_objects_iterator(self, pb):
-        return MissingObjectsIterator(self.source_store, self.source, pb)
-
-    def dfetch(self, stop_revisions):
-        """Import the gist of the ancestry of a particular revision."""
+    def fetch_objects(self, revs, roundtrip):
+        self.source_store.lock_read()
+        try:
+            todo = list(self.missing_revisions(revs))
+            revidmap = {}
+            pb = ui.ui_factory.nested_progress_bar()
+            try:
+                object_generator = MissingObjectsIterator(
+                    self.source_store, self.source, pb)
+                for (old_revid, git_sha) in object_generator.import_revisions(
+                    todo, roundtrip=roundtrip):
+                    try:
+                        self.mapping.revision_id_bzr_to_foreign(old_revid)
+                    except errors.InvalidRevisionId:
+                        self.target_refs[self.mapping.revid_as_refname(old_revid)] = git_sha
+                    if not roundtrip:
+                        new_revid = self.mapping.revision_id_foreign_to_bzr(git_sha)
+                    else:
+                        new_revid = old_revid
+                    revidmap[old_revid] = (git_sha, new_revid)
+                self.target_store.add_objects(object_generator)
+                return revidmap
+            finally:
+                pb.finished()
+        finally:
+            self.source_store.unlock()
 
     def fetch(self, revision_id=None, pb=None, find_ghosts=False,
             fetch_spec=None, mapped_refs=None):
         if not self.mapping.roundtripping:
             raise NoPushSupport()
-        self.source_store.lock_read()
-        try:
-            if mapped_refs is not None:
-                stop_revisions = mapped_refs
-            elif revision_id is not None:
-                stop_revisions = [(None, revision_id)]
-            elif fetch_spec is not None:
-                stop_revisions = [(None, revid) for revid in fetch_spec.heads]
+        if mapped_refs is not None:
+            stop_revisions = mapped_refs
+        elif revision_id is not None:
+            stop_revisions = [(None, revision_id)]
+        elif fetch_spec is not None:
+            recipe = fetch_spec.get_recipe()
+            if recipe[0] in ("search", "proxy-search"):
+                stop_revisions = [(None, revid) for revid in recipe[1]]
             else:
-                stop_revisions = [(None, revid) for revid in self.source.all_revision_ids()]
-            todo = list(self.missing_revisions(stop_revisions))
-            pb = ui.ui_factory.nested_progress_bar()
-            try:
-                object_generator = self._get_missing_objects_iterator(pb)
-                for (revid, git_sha) in object_generator.import_revisions(
-                    todo, roundtrip=True):
-                    try:
-                        self.mapping.revision_id_bzr_to_foreign(revid)
-                    except errors.InvalidRevisionId:
-                        self.target_refs[self.mapping.revid_as_refname(revid)] = git_sha
-                self.target_store.add_objects(object_generator)
-            finally:
-                pb.finished()
-        finally:
-            self.source_store.unlock()
+                raise AssertionError("Unsupported search result type %s" % recipe[0])
+        else:
+            stop_revisions = [(None, revid) for revid in self.source.all_revision_ids()]
+        self.fetch_objects(stop_revisions, roundtrip=True)
 
     @staticmethod
     def is_compatible(source, target):
@@ -307,8 +323,10 @@ class InterToLocalGitRepository(InterToGitRepository):
 
 class InterToRemoteGitRepository(InterToGitRepository):
 
-    def dfetch_refs(self, update_refs):
+    def fetch_refs(self, update_refs, lossy):
         """Import the gist of the ancestry of a particular revision."""
+        if not lossy:
+            raise NoPushSupport()
         unpeel_map = UnpeelMap.from_repository(self.source)
         revidmap = {}
         def determine_wants(old_refs):
@@ -328,11 +346,9 @@ class InterToRemoteGitRepository(InterToGitRepository):
                     self.source_store.generate_lossy_pack_contents)
         finally:
             self.source_store.unlock()
+        # FIXME: revidmap?
         return revidmap, self.old_refs, self.new_refs
 
-    def fetch_refs(self, update_refs):
-        raise NoPushSupport()
-
     @staticmethod
     def is_compatible(source, target):
         """Be compatible with GitRepository."""
