@@ -20,9 +20,11 @@ from bzrlib import (
     trace,
     ui,
     urlutils,
+    version_info as bzrlib_version,
     )
 from bzrlib.errors import (
     BzrError,
+    InProcessTransport,
     InvalidRevisionId,
     NoSuchFile,
     NoSuchRevision,
@@ -46,8 +48,6 @@ from bzrlib.plugins.git.branch import (
 from bzrlib.plugins.git.dir import (
     GitControlDirFormat,
     GitDir,
-    GitLockableFiles,
-    GitLock,
     )
 from bzrlib.plugins.git.errors import (
     GitSmartRemoteNotSupported,
@@ -60,11 +60,12 @@ from bzrlib.plugins.git.repository import (
     GitRepository,
     )
 from bzrlib.plugins.git.refs import (
-    extract_tags,
     branch_name_to_ref,
+    is_peeled,
     )
 
-import dulwich as git
+import dulwich
+import dulwich.client
 from dulwich.errors import (
     GitProtocolError,
     )
@@ -72,6 +73,7 @@ from dulwich.pack import (
     Pack,
     PackData,
     )
+from dulwich.repo import DictRefsContainer
 import os
 import tempfile
 import urllib
@@ -125,6 +127,11 @@ class GitSmartTransport(Transport):
             trace.mutter('host: %r, user: %r, port: %r, path: %r',
                          self._host, self._username, self._port, self._path)
         self._client = _client
+        if "," in self._path and bzrlib_version < (2, 5, 0):
+            trace.warning(
+                "ignoring parameters %r, not supported in bzr < 2.5.",
+                self._path.rsplit(",", 1)[1])
+        self._stripped_path = self._path.rsplit(",", 1)[0]
 
     def external_url(self):
         return self.base
@@ -136,26 +143,7 @@ class GitSmartTransport(Transport):
         raise NotImplementedError(self._get_client)
 
     def _get_path(self):
-        return self._path
-
-    def fetch_pack(self, determine_wants, graph_walker, pack_data, progress=None):
-        if progress is None:
-            def progress(text):
-                trace.info("git: %s" % text)
-        client = self._get_client(thin_packs=False)
-        try:
-            return client.fetch_pack(self._get_path(), determine_wants,
-                graph_walker, pack_data, progress)
-        except GitProtocolError, e:
-            raise parse_git_error(self.external_url(), e)
-
-    def send_pack(self, get_changed_refs, generate_pack_contents):
-        client = self._get_client(thin_packs=False)
-        try:
-            return client.send_pack(self._get_path(), get_changed_refs,
-                generate_pack_contents)
-        except GitProtocolError, e:
-            raise parse_git_error(self.external_url(), e)
+        return self._stripped_path
 
     def get(self, path):
         raise NoSuchFile(path)
@@ -182,7 +170,7 @@ class TCPGitSmartTransport(GitSmartTransport):
             ret = self._client
             self._client = None
             return ret
-        return git.client.TCPGitClient(self._host, self._port,
+        return dulwich.client.TCPGitClient(self._host, self._port,
             thin_packs=thin_packs, report_activity=self._report_activity)
 
 
@@ -191,9 +179,10 @@ class SSHGitSmartTransport(GitSmartTransport):
     _scheme = 'git+ssh'
 
     def _get_path(self):
-        if self._path.startswith("/~/"):
-            return self._path[3:]
-        return self._path
+        path = self._stripped_path
+        if path.startswith("/~/"):
+            return path[3:]
+        return path
 
     def _get_client(self, thin_packs):
         if self._client is not None:
@@ -201,7 +190,7 @@ class SSHGitSmartTransport(GitSmartTransport):
             self._client = None
             return ret
         location_config = config.LocationConfig(self.base)
-        client = git.client.SSHGitClient(self._host, self._port, self._username,
+        client = dulwich.client.SSHGitClient(self._host, self._port, self._username,
             thin_packs=thin_packs, report_activity=self._report_activity)
         # Set up alternate pack program paths
         upload_pack = location_config.get_user_option('git_upload_pack')
@@ -215,31 +204,90 @@ class SSHGitSmartTransport(GitSmartTransport):
 
 class RemoteGitDir(GitDir):
 
-    def __init__(self, transport, lockfiles, format):
+    def __init__(self, transport, format, get_client, client_path):
         self._format = format
         self.root_transport = transport
         self.transport = transport
-        self._lockfiles = lockfiles
         self._mode_check_done = None
+        self._get_client = get_client
+        self._client_path = client_path
+        self.base = self.root_transport.base
+        self._refs = None
+
+    def fetch_pack(self, determine_wants, graph_walker, pack_data, progress=None):
+        if progress is None:
+            def progress(text):
+                trace.info("git: %s" % text)
+        def wrap_determine_wants(refs_dict):
+            return determine_wants(remote_refs_dict_to_container(refs_dict))
+        client = self._get_client(thin_packs=False)
+        try:
+            refs_dict = client.fetch_pack(self._client_path, wrap_determine_wants,
+                graph_walker, pack_data, progress)
+            self._refs = remote_refs_dict_to_container(refs_dict)
+            return refs_dict
+        except GitProtocolError, e:
+            raise parse_git_error(self.transport.external_url(), e)
+
+    def send_pack(self, get_changed_refs, generate_pack_contents):
+        client = self._get_client(thin_packs=False)
+        try:
+            return client.send_pack(self._client_path, get_changed_refs,
+                generate_pack_contents)
+        except GitProtocolError, e:
+            raise parse_git_error(self.transport.external_url(), e)
+
+    def destroy_branch(self, name=None):
+        refname = self._get_selected_ref(name)
+        if refname is None:
+            refname = "HEAD"
+        def get_changed_refs(old_refs):
+            ret = dict(old_refs)
+            if not refname in ret:
+                raise NotBranchError(self.user_url)
+            ret[refname] = "00" * 20
+            return ret
+        self.send_pack(get_changed_refs, lambda have, want: [])
 
     @property
     def user_url(self):
         return self.control_url
 
-    def _branch_name_to_ref(self, name, default=None):
-        return branch_name_to_ref(name, default=default)
+    @property
+    def user_transport(self):
+        return self.root_transport
+
+    @property
+    def control_url(self):
+        return self.control_transport.base
+
+    @property
+    def control_transport(self):
+        return self.root_transport
 
     def open_repository(self):
-        return RemoteGitRepository(self, self._lockfiles)
+        return RemoteGitRepository(self)
 
-    def open_branch(self, name=None, unsupported=False, ignore_fallbacks=False):
+    def open_branch(self, name=None, unsupported=False,
+            ignore_fallbacks=False, ref=None, possible_transports=None):
         repo = self.open_repository()
-        refname = self._branch_name_to_ref(name)
-        return RemoteGitBranch(self, repo, refname, self._lockfiles)
+        refname = self._get_selected_ref(name, ref)
+        return RemoteGitBranch(self, repo, refname)
 
     def open_workingtree(self, recommend_upgrade=False):
         raise NotLocalUrl(self.transport.base)
 
+    def get_peeled(self, name):
+        return self.get_refs_container().get_peeled(name)
+
+    def get_refs_container(self):
+        if self._refs is not None:
+            return self._refs
+        refs_dict = self.fetch_pack(lambda x: [], None,
+            lambda x: None, lambda x: trace.mutter("git: %s" % x))
+        self._refs = remote_refs_dict_to_container(refs_dict)
+        return self._refs
+
 
 class EmptyObjectStoreIterator(dict):
 
@@ -283,6 +331,21 @@ class TemporaryPackIterator(Pack):
             os.remove(self._data_path)
 
 
+class BzrGitHttpClient(dulwich.client.HttpGitClient):
+
+    def __init__(self, transport, *args, **kwargs):
+        self.transport = transport
+        super(BzrGitHttpClient, self).__init__(transport.external_url(), *args, **kwargs)
+        import urllib2
+        self._http_perform = getattr(self.transport, "_perform", urllib2.urlopen)
+
+    def _perform(self, req):
+        req.accepted_errors = (200, 404)
+        req.follow_redirections = True
+        req.redirected_to = None
+        return self._http_perform(req)
+
+
 class RemoteGitControlDirFormat(GitControlDirFormat):
     """The .git directory control format."""
 
@@ -292,6 +355,12 @@ class RemoteGitControlDirFormat(GitControlDirFormat):
     def _known_formats(self):
         return set([RemoteGitControlDirFormat()])
 
+    def is_initializable(self):
+        return False
+
+    def is_supported(self):
+        return True
+
     def open(self, transport, _found=None):
         """Open this directory.
 
@@ -300,12 +369,16 @@ class RemoteGitControlDirFormat(GitControlDirFormat):
         url = transport.base
         if url.startswith('readonly+'):
             url = url[len('readonly+'):]
-        if (not url.startswith("git://") and not url.startswith("git+")):
+        if isinstance(transport, GitSmartTransport):
+            get_client = transport._get_client
+            client_path = transport._get_path()
+        elif urlparse.urlsplit(transport.external_url())[0] in ("http", "https"):
+            def get_client(thin_packs=False):
+                return BzrGitHttpClient(transport, thin_packs=thin_packs)
+            client_path = transport._path
+        else:
             raise NotBranchError(transport.base)
-        if not isinstance(transport, GitSmartTransport):
-            raise NotBranchError(transport.base)
-        lockfiles = GitLockableFiles(transport, GitLock())
-        return RemoteGitDir(transport, lockfiles, self)
+        return RemoteGitDir(transport, self, get_client, client_path)
 
     def get_format_description(self):
         return "Remote Git Repository"
@@ -313,34 +386,33 @@ class RemoteGitControlDirFormat(GitControlDirFormat):
     def initialize_on_transport(self, transport):
         raise UninitializableFormat(self)
 
+    def supports_transport(self, transport):
+        try:
+            external_url = transport.external_url()
+        except InProcessTransport:
+            raise NotBranchError(path=transport.base)
+        return (external_url.startswith("http:") or
+                external_url.startswith("https:") or
+                external_url.startswith("git+") or
+                external_url.startswith("git:"))
+
 
 class RemoteGitRepository(GitRepository):
 
-    def __init__(self, gitdir, lockfiles):
-        GitRepository.__init__(self, gitdir, lockfiles)
-        self._refs = None
-
     @property
     def user_url(self):
         return self.control_url
 
     def get_parent_map(self, revids):
-        raise GitSmartRemoteNotSupported()
-
-    def get_refs(self):
-        if self._refs is not None:
-            return self._refs
-        self._refs = self.bzrdir.root_transport.fetch_pack(lambda x: [], None,
-            lambda x: None, lambda x: trace.mutter("git: %s" % x))
-        return self._refs
+        raise GitSmartRemoteNotSupported(self.get_parent_map, self)
 
     def fetch_pack(self, determine_wants, graph_walker, pack_data,
                    progress=None):
-        return self._transport.fetch_pack(determine_wants, graph_walker,
+        return self.bzrdir.fetch_pack(determine_wants, graph_walker,
                                           pack_data, progress)
 
     def send_pack(self, get_changed_refs, generate_pack_contents):
-        return self._transport.send_pack(get_changed_refs, generate_pack_contents)
+        return self.bzrdir.send_pack(get_changed_refs, generate_pack_contents)
 
     def fetch_objects(self, determine_wants, graph_walker, resolve_ext_ref,
                       progress=None):
@@ -370,16 +442,17 @@ class RemoteGitRepository(GitRepository):
         # Not really an easy way to parse foreign revids here..
         return mapping.revision_id_foreign_to_bzr(foreign_revid)
 
+    def revision_tree(self, revid):
+        raise GitSmartRemoteNotSupported(self.revision_tree, self)
+
+    def get_revisions(self, revids):
+        raise GitSmartRemoteNotSupported(self.get_revisions, self)
+
 
 class RemoteGitTagDict(GitTags):
 
-    def get_refs(self):
-        return self.repository.get_refs()
-
-    def _iter_tag_refs(self, refs):
-        for k, (peeled, unpeeled) in extract_tags(refs).iteritems():
-            yield (k, peeled, unpeeled,
-                  self.branch.mapping.revision_id_foreign_to_bzr(peeled))
+    def get_refs_container(self):
+        return self.repository.bzrdir.get_refs_container()
 
     def set_tag(self, name, revid):
         # FIXME: Not supported yet, should do a push of a new ref
@@ -388,10 +461,12 @@ class RemoteGitTagDict(GitTags):
 
 class RemoteGitBranch(GitBranch):
 
-    def __init__(self, bzrdir, repository, name, lockfiles):
+    def __init__(self, bzrdir, repository, name):
         self._sha = None
-        super(RemoteGitBranch, self).__init__(bzrdir, repository, name,
-                lockfiles)
+        super(RemoteGitBranch, self).__init__(bzrdir, repository, name)
+
+    def last_revision_info(self):
+        raise GitSmartRemoteNotSupported(self.last_revision_info, self)
 
     @property
     def user_url(self):
@@ -402,29 +477,24 @@ class RemoteGitBranch(GitBranch):
         return self.base
 
     def revision_history(self):
-        raise GitSmartRemoteNotSupported()
+        raise GitSmartRemoteNotSupported(self.revision_history, self)
+
+    def revision_id_to_revno(self, revision_id):
+        raise GitSmartRemoteNotSupported(self.revision_id_to_revno, self)
 
     def last_revision(self):
         return self.lookup_foreign_revision_id(self.head)
 
-    def _get_config(self):
-        class EmptyConfig(object):
-
-            def _get_configobj(self):
-                return config.ConfigObj()
-
-        return EmptyConfig()
-
     @property
     def head(self):
         if self._sha is not None:
             return self._sha
-        heads = self.repository.get_refs()
-        name = self.bzrdir._branch_name_to_ref(self.name, "HEAD")
-        if name in heads:
-            self._sha = heads[name]
-        else:
-            raise NoSuchRef(self.name)
+        refs = self.bzrdir.get_refs_container()
+        name = branch_name_to_ref(self.name, "HEAD")
+        try:
+            self._sha = refs[name]
+        except KeyError:
+            raise NoSuchRef(name, self.repository.user_url, refs)
         return self._sha
 
     def _synchronize_history(self, destination, revision_id):
@@ -436,3 +506,17 @@ class RemoteGitBranch(GitBranch):
 
     def set_push_location(self, url):
         pass
+
+
+def remote_refs_dict_to_container(refs_dict):
+    base = {}
+    peeled = {}
+    for k, v in refs_dict.iteritems():
+        if is_peeled(k):
+            peeled[k[:-3]] = v
+        else:
+            base[k] = v
+            peeled[k] = v
+    ret = DictRefsContainer(base)
+    ret._peeled = peeled
+    return ret
