@@ -24,7 +24,10 @@ from bzrlib import (
     inventory,
     repository,
     revision,
+    transactions,
+    version_info as bzrlib_version,
     )
+from bzrlib.decorators import only_raises
 try:
     from bzrlib.revisiontree import InventoryRevisionTree
 except ImportError: # bzr < 2.4
@@ -36,6 +39,9 @@ from bzrlib.foreign import (
 from bzrlib.plugins.git.commit import (
     GitCommitBuilder,
     )
+from bzrlib.plugins.git.errors import (
+    NotCommitError,
+    )
 from bzrlib.plugins.git.filegraph import (
     GitFileLastChangeScanner,
     GitFileParentProvider,
@@ -101,6 +107,22 @@ class GitCheck(check.Check):
         pass
 
 
+_optimisers_loaded = False
+
+def lazy_load_optimisers():
+    global _optimisers_loaded
+    if _optimisers_loaded:
+        return
+    from bzrlib.plugins.git import fetch, push
+    for optimiser in [fetch.InterRemoteGitNonGitRepository,
+                      fetch.InterLocalGitNonGitRepository,
+                      fetch.InterGitGitRepository,
+                      push.InterToLocalGitRepository,
+                      push.InterToRemoteGitRepository]:
+        repository.InterRepository.register_optimiser(optimiser)
+    _optimisers_loaded = True
+
+
 class GitRepository(ForeignRepository):
     """An adapter to git repositories for bzr."""
 
@@ -108,22 +130,88 @@ class GitRepository(ForeignRepository):
     vcs = foreign_vcs_git
     chk_bytes = None
 
-    def __init__(self, gitdir, lockfiles):
-        ForeignRepository.__init__(self, GitRepositoryFormat(), gitdir, lockfiles)
-        from bzrlib.plugins.git import fetch, push
-        for optimiser in [fetch.InterRemoteGitNonGitRepository,
-                          fetch.InterLocalGitNonGitRepository,
-                          fetch.InterGitGitRepository,
-                          push.InterToLocalGitRepository,
-                          push.InterToRemoteGitRepository]:
-            repository.InterRepository.register_optimiser(optimiser)
+    def __init__(self, gitdir):
+        if bzrlib_version >= (2, 5):
+            control_files = None
+        else:
+            class DummyControlFiles(object):
+                def __init__(self):
+                    self._transport = gitdir.root_transport
+            control_files = DummyControlFiles()
+        self._transport = gitdir.root_transport
+        super(GitRepository, self).__init__(GitRepositoryFormat(),
+            gitdir, control_files)
+        self.base = gitdir.root_transport.base
+        lazy_load_optimisers()
+        self._lock_mode = None
+        self._lock_count = 0
 
     def add_fallback_repository(self, basis_url):
-        raise errors.UnstackableRepositoryFormat(self._format, self.control_transport.base)
+        raise errors.UnstackableRepositoryFormat(self._format,
+            self.control_transport.base)
 
     def is_shared(self):
         return False
 
+    def get_physical_lock_status(self):
+        return False
+
+    def lock_write(self):
+        """See Branch.lock_write()."""
+        if self._lock_mode:
+            assert self._lock_mode == 'w'
+            self._lock_count += 1
+        else:
+            self._lock_mode = 'w'
+            self._lock_count = 1
+        return GitRepositoryLock(self)
+
+    def break_lock(self):
+        raise NotImplementedError(self.break_lock)
+
+    def dont_leave_lock_in_place(self):
+        raise NotImplementedError(self.dont_leave_lock_in_place)
+
+    def leave_lock_in_place(self):
+        raise NotImplementedError(self.leave_lock_in_place)
+
+    def lock_read(self):
+        if self._lock_mode:
+            assert self._lock_mode in ('r', 'w')
+            self._lock_count += 1
+        else:
+            self._lock_mode = 'r'
+            self._lock_count = 1
+        return self
+
+    @only_raises(errors.LockNotHeld, errors.LockBroken)
+    def unlock(self):
+        if self._lock_count == 0:
+            raise errors.LockNotHeld(self)
+        if self._lock_count == 1 and self._lock_mode == 'w':
+            if self._write_group is not None:
+                self.abort_write_group()
+                self._lock_count -= 1
+                self._lock_mode = None
+                raise errors.BzrError(
+                    'Must end write groups before releasing write locks.')
+        self._lock_count -= 1
+        if self._lock_count == 0:
+            self._lock_mode = None
+
+    def is_write_locked(self):
+        return (self._lock_mode == 'w')
+
+    def is_locked(self):
+        return (self._lock_mode is not None)
+
+    def get_transaction(self):
+        """See Repository.get_transaction()."""
+        if self._write_group is None:
+            return transactions.PassThroughTransaction()
+        else:
+            return self._write_group
+
     def reconcile(self, other=None, thorough=False):
         """Reconcile this repository."""
         reconciler = RepoReconciler(self, thorough=thorough)
@@ -146,20 +234,29 @@ class GitRepository(ForeignRepository):
     def revision_graph_can_have_wrong_parents(self):
         return False
 
-    def dfetch(self, source, stop_revision):
-        interrepo = repository.InterRepository.get(source, self)
-        return interrepo.dfetch(stop_revision)
-
     def add_signature_text(self, revid, signature):
         raise errors.UnsupportedOperation(self.add_signature_text, self)
 
+    def sign_revision(self, revision_id, gpg_strategy):
+        raise errors.UnsupportedOperation(self.add_signature_text, self)
+
+
+class GitRepositoryLock(object):
+    """Subversion lock."""
+
+    def __init__(self, repository):
+        self.repository_token = None
+        self.repository = repository
+
+    def unlock(self):
+        self.repository.unlock()
+
 
 class LocalGitRepository(GitRepository):
     """Git repository on the file system."""
 
-    def __init__(self, gitdir, lockfiles):
-        GitRepository.__init__(self, gitdir, lockfiles)
-        self.base = gitdir.root_transport.base
+    def __init__(self, gitdir):
+        GitRepository.__init__(self, gitdir)
         self._git = gitdir._git
         self._file_change_scanner = GitFileLastChangeScanner(self)
 
@@ -185,7 +282,8 @@ class LocalGitRepository(GitRepository):
             lossy)
 
     def get_file_graph(self):
-        return _mod_graph.Graph(GitFileParentProvider(self._file_change_scanner))
+        return _mod_graph.Graph(GitFileParentProvider(
+            self._file_change_scanner))
 
     def iter_files_bytes(self, desired_files):
         """Iterate through file versions.
@@ -208,7 +306,8 @@ class LocalGitRepository(GitRepository):
         """
         per_revision = {}
         for (file_id, revision_id, identifier) in desired_files:
-            per_revision.setdefault(revision_id, []).append((file_id, identifier))
+            per_revision.setdefault(revision_id, []).append(
+                (file_id, identifier))
         for revid, files in per_revision.iteritems():
             (commit_id, mapping) = self.lookup_bzr_revision_id(revid)
             try:
@@ -234,7 +333,6 @@ class LocalGitRepository(GitRepository):
                     else:
                         raise AssertionError("file text resolved to %r" % obj)
 
-
     def _iter_revision_ids(self):
         mapping = self.get_mapping()
         for sha in self._git.object_store:
@@ -248,27 +346,37 @@ class LocalGitRepository(GitRepository):
     def all_revision_ids(self):
         ret = set([])
         for git_sha, revid, roundtrip_revid in self._iter_revision_ids():
-            ret.add(revid)
             if roundtrip_revid:
                 ret.add(roundtrip_revid)
+            else:
+                ret.add(revid)
         return ret
 
+    def _get_parents(self, revid):
+        if type(revid) != str:
+            raise ValueError
+        try:
+            (hexsha, mapping) = self.lookup_bzr_revision_id(revid)
+        except errors.NoSuchRevision:
+            return None
+        try:
+            commit = self._git[hexsha]
+        except KeyError:
+            return None
+        return [
+            self.lookup_foreign_revision_id(p, mapping)
+            for p in commit.parents]
+
     def get_parent_map(self, revids):
         parent_map = {}
         for revision_id in revids:
-            assert isinstance(revision_id, str)
+            parents = self._get_parents(revision_id)
             if revision_id == revision.NULL_REVISION:
                 parent_map[revision_id] = ()
                 continue
-            hexsha, mapping = self.lookup_bzr_revision_id(revision_id)
-            try:
-                commit = self._git[hexsha]
-            except KeyError:
+            if parents is None:
                 continue
-            parents = [
-                self.lookup_foreign_revision_id(p, mapping)
-                for p in commit.parents]
-            if parents == []:
+            if len(parents) == 0:
                 parents = [revision.NULL_REVISION]
             parent_map[revision_id] = tuple(parents)
         return parent_map
@@ -279,7 +387,13 @@ class LocalGitRepository(GitRepository):
         pending = set(revision_ids)
         parent_map = {}
         while pending:
-            this_parent_map = self.get_parent_map(pending)
+            this_parent_map = {}
+            for revid in pending:
+                if revid == revision.NULL_REVISION:
+                    continue
+                parents = self._get_parents(revid)
+                if parents is not None:
+                    this_parent_map[revid] = parents
             parent_map.update(this_parent_map)
             pending = set()
             map(pending.update, this_parent_map.itervalues())
@@ -309,6 +423,8 @@ class LocalGitRepository(GitRepository):
         commit = self._git.object_store[foreign_revid]
         while isinstance(commit, Tag):
             commit = self._git[commit.object[1]]
+        if not isinstance(commit, Commit):
+            raise NotCommitError(commit.id)
         rev, roundtrip_revid, verifiers = mapping.import_commit(commit,
             mapping.revision_id_foreign_to_bzr)
         # FIXME: check testament before doing this?
@@ -318,16 +434,28 @@ class LocalGitRepository(GitRepository):
             return rev.revision_id
 
     def has_signature_for_revision_id(self, revision_id):
+        """Check whether a GPG signature is present for this revision.
+
+        This is never the case for Git repositories.
+        """
         return False
 
     def lookup_bzr_revision_id(self, bzr_revid, mapping=None):
+        """Lookup a bzr revision id in a Git repository.
+
+        :param bzr_revid: Bazaar revision id
+        :param mapping: Optional mapping to use
+        :return: Tuple with git commit id, mapping that was used and supplement
+            details
+        """
         try:
-            return mapping_registry.revision_id_bzr_to_foreign(bzr_revid)
+            (git_sha, mapping) = mapping_registry.revision_id_bzr_to_foreign(bzr_revid)
         except errors.InvalidRevisionId:
             if mapping is None:
                 mapping = self.get_mapping()
             try:
-                return (self._git.refs[mapping.revid_as_refname(bzr_revid)], mapping)
+                return (self._git.refs[mapping.revid_as_refname(bzr_revid)],
+                        mapping)
             except KeyError:
                 # Update refs from Git commit objects
                 # FIXME: Hitting this a lot will be very inefficient...
@@ -339,6 +467,8 @@ class LocalGitRepository(GitRepository):
                     if roundtrip_revid == bzr_revid:
                         return git_sha, mapping
                 raise errors.NoSuchRevision(self, bzr_revid)
+        else:
+            return (git_sha, mapping)
 
     def get_revision(self, revision_id):
         if not isinstance(revision_id, str):
@@ -392,8 +522,8 @@ class LocalGitRepository(GitRepository):
         raise NotImplementedError(self.get_inventory)
 
     def set_make_working_trees(self, trees):
+        raise errors.UnsupportedOperation(self.set_make_working_trees, self)
         # TODO: Set bare= in the configuration bug=777065
-        raise NotImplementedError(self.set_make_working_trees)
 
     def fetch_objects(self, determine_wants, graph_walker, resolve_ext_ref,
         progress=None):
@@ -412,7 +542,9 @@ class GitRepositoryFormat(repository.RepositoryFormat):
     supports_external_lookups = False
     supports_full_versioned_files = False
     supports_revision_signatures = False
+    supports_nesting_repositories = False
     revision_graph_can_have_wrong_parents = False
+    supports_unreferenced_revisions = True
 
     @property
     def _matchingbzrdir(self):
