@@ -28,10 +28,13 @@ from dulwich.repo import (
     )
 
 import os
+import urllib
 
 from bzrlib import (
     errors,
     revision,
+    urlutils,
+    version_info as bzrlib_version,
     )
 from bzrlib.branch import (
     Branch,
@@ -43,12 +46,18 @@ from bzrlib.bzrdir import (
 from bzrlib.repository import (
     Repository,
     )
+from bzrlib.symbol_versioning import (
+    deprecated_in,
+    )
+from bzrlib.tests import TestSkipped
 
 from bzrlib.plugins.git import (
-    LocalGitControlDirFormat,
     branch,
     tests,
     )
+from bzrlib.plugins.git.dir import (
+    LocalGitControlDirFormat,
+    )
 from bzrlib.plugins.git.mapping import (
     default_mapping,
     )
@@ -56,6 +65,19 @@ from bzrlib.plugins.git.mapping import (
 
 class TestGitBranch(tests.TestCaseInTempDir):
 
+    def test_open_by_ref(self):
+        GitRepo.init('.')
+        url = "%s,ref=%s" % (
+            urlutils.local_path_to_url(self.test_dir),
+            urllib.quote("refs/remotes/origin/unstable", safe='')
+            )
+        if bzrlib_version < (2, 5, 0):
+            self.assertRaises(errors.NotBranchError, BzrDir.open, url)
+            raise TestSkipped("opening by ref not supported with bzr < 2.5")
+        d = BzrDir.open(url)
+        b = d.create_branch()
+        self.assertEquals(b.ref, "refs/remotes/origin/unstable")
+
     def test_open_existing(self):
         GitRepo.init('.')
         d = BzrDir.open('.')
@@ -66,7 +88,10 @@ class TestGitBranch(tests.TestCaseInTempDir):
         GitRepo.init('.')
         d = BzrDir.open('.')
         thebranch = d.create_branch()
-        self.assertEquals("<LocalGitBranch('file://%s/', 'HEAD')>" % self.test_dir, repr(thebranch))
+        self.assertEquals(
+            "<LocalGitBranch('%s/', u'master')>" % (
+                urlutils.local_path_to_url(self.test_dir),),
+            repr(thebranch))
 
     def test_last_revision_is_null(self):
         GitRepo.init('.')
@@ -96,8 +121,13 @@ class TestGitBranch(tests.TestCaseInTempDir):
         revb = r.do_commit("b", committer="Somebody <foo@example.com>")
 
         thebranch = Branch.open('.')
+        (warnings, history) = self.callCatchWarnings(thebranch.revision_history)
+        self.assertTrue(
+            warnings == [] or 
+            (len(warnings) == 1 and isinstance(warnings[0], DeprecationWarning)),
+            warnings)
         self.assertEqual([default_mapping.revision_id_foreign_to_bzr(r) for r in (reva, revb)],
-                         thebranch.revision_history())
+                         history)
 
     def test_tag_annotated(self):
         reva = self.simple_commit_a()
@@ -171,7 +201,7 @@ class BranchTests(tests.TestCaseInTempDir):
         mark = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
         gitsha = bb.finish()[mark]
         os.chdir("..")
-        return "d", gitsha
+        return os.path.abspath("d"), gitsha
 
     def make_tworev_branch(self):
         os.mkdir("d")
@@ -194,6 +224,7 @@ class BranchTests(tests.TestCaseInTempDir):
         path, gitsha = self.make_onerev_branch()
         oldrepo = Repository.open(path)
         revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
+        self.assertEquals(gitsha, oldrepo._git.get_refs()["refs/heads/master"])
         newbranch = self.clone_git_branch(path, "f")
         self.assertEquals([revid], newbranch.repository.all_revision_ids())
 
@@ -236,17 +267,20 @@ class BranchTests(tests.TestCaseInTempDir):
         inter_branch.pull(stop_revision=revid1)
         self.assertEquals(revid1, newbranch.last_revision())
 
-    def test_interbranch_limited_pull(self):
+    def test_interbranch_pull_with_tags(self):
         path, (gitsha1, gitsha2) = self.make_tworev_branch()
+        gitrepo = GitRepo(path)
+        gitrepo.refs["refs/tags/sometag"] = gitsha2
         oldrepo = Repository.open(path)
         revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
         revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
         newbranch = self.make_branch('g')
-        inter_branch = InterBranch.get(Branch.open(path), newbranch)
-        inter_branch.pull(limit=1)
+        source_branch = Branch.open(path)
+        source_branch.get_config().set_user_option("branch.fetch_tags", True)
+        inter_branch = InterBranch.get(source_branch, newbranch)
+        inter_branch.pull(stop_revision=revid1)
         self.assertEquals(revid1, newbranch.last_revision())
-        inter_branch.pull(limit=1)
-        self.assertEquals(revid2, newbranch.last_revision())
+        self.assertTrue(newbranch.repository.has_revision(revid2))
 
 
 class ForeignTestsBranchFactory(object):
@@ -256,6 +290,3 @@ class ForeignTestsBranchFactory(object):
         return d.create_branch()
 
     make_branch = make_empty_branch
-
-
-
