@@ -18,6 +18,8 @@
 
 from cStringIO import StringIO
 
+import os
+
 from dulwich.errors import (
     NotGitRepository,
     NoIndexPresent,
@@ -52,6 +54,9 @@ from dulwich.repo import (
     write_packed_refs,
     )
 
+from bzrlib import (
+    transport as _mod_transport,
+    )
 from bzrlib.errors import (
     FileExists,
     NoSuchFile,
@@ -93,7 +98,11 @@ class TransportRefsContainer(RefsContainer):
 
     def allkeys(self):
         keys = set()
-        if self.transport.has("HEAD"):
+        try:
+            self.transport.get_bytes("HEAD")
+        except NoSuchFile:
+            pass
+        else:
             keys.add("HEAD")
         try:
             iter_files = list(self.transport.clone("refs").iter_files_recursive())
@@ -276,26 +285,31 @@ class TransportRefsContainer(RefsContainer):
         self._remove_packed_ref(name)
         return True
 
+    def get(self, name, default=None):
+        try:
+            return self[name]
+        except KeyError:
+            return default
+
 
 class TransportRepo(BaseRepo):
 
-    def __init__(self, transport):
+    def __init__(self, transport, bare, refs_text=None):
         self.transport = transport
-        try:
-            if self.transport.has(".git/%s" % OBJECTDIR):
-                self.bare = False
-                self._controltransport = self.transport.clone('.git')
-            elif self.transport.has_any(["info/refs", OBJECTDIR, REFSDIR]):
-                self.bare = True
-                self._controltransport = self.transport
-            else:
-                raise NotGitRepository(self.transport)
-        except NoSuchFile:
-            raise NotGitRepository(self.transport)
+        self.bare = bare
+        if self.bare:
+            self._controltransport = self.transport
+        else:
+            self._controltransport = self.transport.clone('.git')
         object_store = TransportObjectStore(
             self._controltransport.clone(OBJECTDIR))
+        if refs_text is not None:
+            from dulwich.repo import InfoRefsContainer # dulwich >= 0.8.2
+            refs_container = InfoRefsContainer(StringIO(refs_text))
+        else:
+            refs_container = TransportRefsContainer(self._controltransport)
         super(TransportRepo, self).__init__(object_store, 
-                TransportRefsContainer(self._controltransport))
+                refs_container)
 
     def get_named_file(self, path):
         """Get a file from the control dir with a specific name.
@@ -346,7 +360,7 @@ class TransportRepo(BaseRepo):
             control_transport.mkdir("/".join(d))
         control_transport.mkdir(OBJECTDIR)
         TransportObjectStore.init(control_transport.clone(OBJECTDIR))
-        ret = cls(transport)
+        ret = cls(transport, bare)
         ret.refs.set_symbolic_ref("HEAD", "refs/heads/master")
         ret._init_files(bare)
         return ret
@@ -363,6 +377,7 @@ class TransportObjectStore(PackBasedObjectStore):
         super(TransportObjectStore, self).__init__()
         self.transport = transport
         self.pack_transport = self.transport.clone(PACKDIR)
+        self._alternates = None
 
     def __repr__(self):
         return "%s(%r)" % (self.__class__.__name__, self.transport)
@@ -370,6 +385,35 @@ class TransportObjectStore(PackBasedObjectStore):
     def _pack_cache_stale(self):
         return False # FIXME
 
+    @property
+    def alternates(self):
+        if self._alternates is not None:
+            return self._alternates
+        self._alternates = []
+        for path in self._read_alternate_paths():
+            # FIXME: Check path
+            t = _mod_transport.get_transport_from_path(path)
+            self._alternates.append(self.__class__(t))
+        return self._alternates
+
+    def _read_alternate_paths(self):
+        try:
+            f = self.transport.get("info/alternates")
+        except NoSuchFile:
+            return []
+        ret = []
+        try:
+            for l in f.readlines():
+                l = l.rstrip("\n")
+                if l[0] == "#":
+                    continue
+                if os.path.isabs(l):
+                    continue
+                ret.append(l)
+            return ret
+        finally:
+            f.close()
+
     def _pack_names(self):
         try:
             f = self.transport.get('info/packs')
@@ -515,8 +559,6 @@ class TransportObjectStore(PackBasedObjectStore):
         self._add_known_pack(final_pack)
         return final_pack
 
-
-
     def add_pack(self):
         """Add a new pack to this object store. 
 
