@@ -34,7 +34,6 @@ from dulwich.objects import (
     ZERO_SHA,
     )
 import os
-import posix
 import posixpath
 import stat
 import sys
@@ -44,8 +43,7 @@ from bzrlib import (
     conflicts as _mod_conflicts,
     ignores,
     inventory,
-    lockable_files,
-    lockdir,
+    lock,
     osutils,
     trace,
     transport,
@@ -55,6 +53,7 @@ from bzrlib import (
 from bzrlib.decorators import (
     needs_read_lock,
     )
+from bzrlib.mutabletree import needs_tree_write_lock
 
 
 from bzrlib.plugins.git.dir import (
@@ -83,17 +82,6 @@ class GitWorkingTree(workingtree.WorkingTree):
         self.mapping = self.repository.get_mapping()
         self._branch = branch
         self._transport = bzrdir.transport
-
-        self.controldir = self.bzrdir.transport.local_abspath('bzr')
-
-        try:
-            os.makedirs(self.controldir)
-            os.makedirs(os.path.join(self.controldir, 'lock'))
-        except OSError:
-            pass
-
-        self._control_files = lockable_files.LockableFiles(
-            transport.get_transport(self.controldir), 'lock', lockdir.LockDir)
         self._format = GitWorkingTreeFormat()
         self.index = index
         self._versioned_dirs = None
@@ -102,13 +90,77 @@ class GitWorkingTree(workingtree.WorkingTree):
         self._detect_case_handling()
         self._reset_data()
         self._fileid_map = self._basis_fileid_map.copy()
+        self._lock_mode = None
+        self._lock_count = 0
+
+    def lock_read(self):
+        """Lock the repository for read operations.
+
+        :return: A bzrlib.lock.LogicalLockResult.
+        """
+        if not self._lock_mode:
+            self._lock_mode = 'r'
+            self._lock_count = 1
+        else:
+            self._lock_count += 1
+        self.branch.lock_read()
+        return lock.LogicalLockResult(self.unlock)
+
+    def lock_tree_write(self):
+        if not self._lock_mode:
+            self._lock_mode = 'w'
+            self._lock_count = 1
+        elif self._lock_mode == 'r':
+            raise errors.ReadOnlyError(self)
+        else:
+            self._lock_count +=1
+        self.branch.lock_read()
+        return lock.LogicalLockResult(self.unlock)
+
+    def lock_write(self, token=None):
+        if not self._lock_mode:
+            self._lock_mode = 'w'
+            self._lock_count = 1
+        elif self._lock_mode == 'r':
+            raise errors.ReadOnlyError(self)
+        else:
+            self._lock_count +=1
+        self.branch.lock_write()
+        return lock.LogicalLockResult(self.unlock)
+
+    def is_locked(self):
+        return self._lock_count >= 1
+
+    def get_physical_lock_status(self):
+        return False
+
+    def unlock(self):
+        if not self._lock_count:
+            return lock.cant_unlock_not_held(self)
+        self._cleanup()
+        self._lock_count -= 1
+        if self._lock_count > 0:
+            return
+        self._lock_mode = None
+        self.branch.unlock()
+
+    def _detect_case_handling(self):
+        try:
+            self._transport.stat(".git/cOnFiG")
+        except errors.NoSuchFile:
+            self.case_sensitive = True
+        else:
+            self.case_sensitive = False
+
+    def merge_modified(self):
+        return {}
 
     def set_parent_trees(self, parents_list, allow_leftmost_as_ghost=False):
         self.set_parent_ids([p for p, t in parents_list])
 
     def _index_add_entry(self, path, file_id, kind):
         assert isinstance(path, basestring)
-        assert type(file_id) == str
+        assert type(file_id) == str or file_id is None
         if kind == "directory":
             # Git indexes don't contain directories
             return
@@ -119,8 +171,8 @@ class GitWorkingTree(workingtree.WorkingTree):
             except (errors.NoSuchFile, IOError):
                 # TODO: Rather than come up with something here, use the old index
                 file = StringIO()
-                from posix import stat_result
-                stat_val = stat_result((stat.S_IFREG | 0644, 0, 0, 0, 0, 0, 0, 0, 0, 0))
+                stat_val = os.stat_result(
+                    (stat.S_IFREG | 0644, 0, 0, 0, 0, 0, 0, 0, 0, 0))
             blob.set_raw_string(file.read())
         elif kind == "symlink":
             blob = Blob()
@@ -129,9 +181,10 @@ class GitWorkingTree(workingtree.WorkingTree):
             except (errors.NoSuchFile, OSError):
                 # TODO: Rather than come up with something here, use the 
                 # old index
-                from posix import stat_result
-                stat_val = stat_result((stat.S_IFLNK, 0, 0, 0, 0, 0, 0, 0, 0, 0))
-            blob.set_raw_string(self.get_symlink_target(file_id).encode("utf-8"))
+                stat_val = os.stat_result(
+                    (stat.S_IFLNK, 0, 0, 0, 0, 0, 0, 0, 0, 0))
+            blob.set_raw_string(
+                self.get_symlink_target(file_id, path).encode("utf-8"))
         else:
             raise AssertionError("unknown kind '%s'" % kind)
         # Add object to the repository if it didn't exist yet
@@ -170,15 +223,18 @@ class GitWorkingTree(workingtree.WorkingTree):
                     del self.index[p]
         # FIXME: remove empty directories
 
+    @needs_tree_write_lock
     def unversion(self, file_ids):
         for file_id in file_ids:
             path = self.id2path(file_id)
             self._unversion_path(path)
+        self.flush()
 
     def check_state(self):
         """Check that the working state is/isn't valid."""
         pass
 
+    @needs_tree_write_lock
     def remove(self, files, verbose=False, to_file=None, keep_files=True,
         force=False):
         """Remove nominated files from the working tree metadata.
@@ -246,6 +302,7 @@ class GitWorkingTree(workingtree.WorkingTree):
             # print only one message (if any) per file.
             if message is not None:
                 trace.note(message)
+        self.flush()
 
     def _add(self, files, ids, kinds):
         for (path, file_id, kind) in zip(files, ids, kinds):
@@ -255,19 +312,20 @@ class GitWorkingTree(workingtree.WorkingTree):
                 file_id = self._fileid_map.lookup_file_id(path.encode("utf-8"))
             self._index_add_entry(path, file_id, kind)
 
+    @needs_tree_write_lock
     def smart_add(self, file_list, recurse=True, action=None, save=True):
         added = []
         ignored = {}
         user_dirs = []
         for filepath in osutils.canonical_relpaths(self.basedir, file_list):
-            if action is not None:
-                file_id = action()
-            else:
-                file_id = None
             abspath = self.abspath(filepath)
             kind = osutils.file_kind(abspath)
+            if action is not None:
+                file_id = action(self, None, filepath, kind)
+            else:
+                file_id = None
             if kind in ("file", "symlink"):
-                if not save:
+                if save:
                     self._index_add_entry(filepath, file_id, kind)
                 added.append(filepath)
             elif kind == "directory":
@@ -279,7 +337,7 @@ class GitWorkingTree(workingtree.WorkingTree):
             abs_user_dir = self.abspath(user_dir)
             for name in os.listdir(abs_user_dir):
                 subp = os.path.join(user_dir, name)
-                if self.is_control_filename(subp):
+                if self.is_control_filename(subp) or self.mapping.is_special_file(subp):
                     continue
                 ignore_glob = self.is_ignored(subp)
                 if ignore_glob is not None:
@@ -291,16 +349,19 @@ class GitWorkingTree(workingtree.WorkingTree):
                     user_dirs.append(subp)
                 else:
                     if action is not None:
-                        file_id = action()
+                        file_id = action(self, None, filepath, kind)
                     else:
                         file_id = None
-                    if not save:
+                    if save:
                         self._index_add_entry(subp, file_id, kind)
+        if added and save:
+            self.flush()
         return added, ignored
 
     def _set_root_id(self, file_id):
         self._fileid_map.set_file_id("", file_id)
 
+    @needs_tree_write_lock
     def move(self, from_paths, to_dir=None, after=False):
         rename_tuples = []
         to_abs = self.abspath(to_dir)
@@ -313,11 +374,11 @@ class GitWorkingTree(workingtree.WorkingTree):
             to_rel = os.path.join(to_dir, from_tail)
             self.rename_one(from_rel, to_rel, after=after)
             rename_tuples.append((from_rel, to_rel))
+        self.flush()
         return rename_tuples
 
+    @needs_tree_write_lock
     def rename_one(self, from_rel, to_rel, after=False):
-        if not after:
-            os.rename(self.abspath(from_rel), self.abspath(to_rel))
         from_path = from_rel.encode("utf-8")
         to_path = to_rel.encode("utf-8")
         if not self.has_filename(to_rel):
@@ -326,13 +387,18 @@ class GitWorkingTree(workingtree.WorkingTree):
         if not from_path in self.index:
             raise errors.BzrMoveFailedError(from_rel, to_rel,
                 errors.NotVersionedError(path=from_rel))
+        if not after:
+            os.rename(self.abspath(from_rel), self.abspath(to_rel))
         self.index[to_path] = self.index[from_path]
         del self.index[from_path]
+        self.flush()
 
     def get_root_id(self):
         return self.path2id("")
 
     def _has_dir(self, path):
+        if path == "":
+            return True
         if self._versioned_dirs is None:
             self._load_dirs()
         return path in self._versioned_dirs
@@ -344,32 +410,25 @@ class GitWorkingTree(workingtree.WorkingTree):
             return self._fileid_map.lookup_file_id(encoded_path)
         return None
 
-    def extras(self):
-        """Yield all unversioned files in this WorkingTree.
-        """
-        present_files = set()
-        for (dirpath, dirnames, filenames) in os.walk(self.basedir):
+    def _iter_files_recursive(self, from_dir=None):
+        if from_dir is None:
+            from_dir = ""
+        for (dirpath, dirnames, filenames) in os.walk(self.abspath(from_dir)):
             dir_relpath = dirpath[len(self.basedir):].strip("/")
             if self.bzrdir.is_control_filename(dir_relpath):
                 continue
             for filename in filenames:
-                relpath = os.path.join(dir_relpath, filename)
-                present_files.add(relpath)
-        return present_files - set(self.index)
+                if not self.mapping.is_special_file(filename):
+                    yield os.path.join(dir_relpath, filename)
 
-    def unlock(self):
-        # non-implementation specific cleanup
-        self._cleanup()
-
-        # reverse order of locking.
-        try:
-            return self._control_files.unlock()
-        finally:
-            self.branch.unlock()
+    def extras(self):
+        """Yield all unversioned files in this WorkingTree.
+        """
+        return set(self._iter_files_recursive()) - set(self.index)
 
     def flush(self):
         # TODO: Maybe this should only write on dirty ?
-        if self._control_files._lock_mode != 'w':
+        if self._lock_mode != 'w':
             raise errors.NotWriteLocked(self)
         self.index.write()
 
@@ -413,8 +472,7 @@ class GitWorkingTree(workingtree.WorkingTree):
             return True
 
     def id2path(self, file_id):
-        if type(file_id) != str:
-            raise AssertionError
+        file_id = osutils.safe_utf8(file_id)
         path = self._fileid_map.lookup_path(file_id)
         # FIXME: What about directories?
         if self._is_versioned(path):
@@ -453,12 +511,13 @@ class GitWorkingTree(workingtree.WorkingTree):
         try:
             head = self.repository._git.head()
         except KeyError, name:
-            raise errors.NotBranchError("branch %s at %s" % (name, self.repository.base))
+            raise errors.NotBranchError("branch %s at %s" % (name,
+                self.repository.base))
         if head == ZERO_SHA:
             self._basis_fileid_map = GitFileIdMap({}, self.mapping)
         else:
-            self._basis_fileid_map = self.mapping.get_fileid_map(self.store.__getitem__,
-                self.store[head].tree)
+            self._basis_fileid_map = self.mapping.get_fileid_map(
+                self.store.__getitem__, self.store[head].tree)
 
     @needs_read_lock
     def get_file_verifier(self, file_id, path=None, stat_value=None):
@@ -485,7 +544,7 @@ class GitWorkingTree(workingtree.WorkingTree):
         return (path in self.index or self._has_dir(path))
 
     def filter_unversioned_files(self, files):
-        return set([p for p in files if self._is_versioned(p.encode("utf-8"))])
+        return set([p for p in files if not self._is_versioned(p.encode("utf-8"))])
 
     def _get_dir_ie(self, path, parent_id):
         file_id = self.path2id(path)
@@ -503,7 +562,8 @@ class GitWorkingTree(workingtree.WorkingTree):
         ret.append((path, ie))
         return ret
 
-    def _get_file_ie(self, path, value, parent_id):
+    def _get_file_ie(self, name, path, value, parent_id):
+        assert isinstance(name, unicode)
         assert isinstance(path, unicode)
         assert isinstance(value, tuple) and len(value) == 10
         (ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags) = value
@@ -511,8 +571,7 @@ class GitWorkingTree(workingtree.WorkingTree):
         if type(file_id) != str:
             raise AssertionError
         kind = mode_kind(mode)
-        ie = inventory.entry_factory[kind](file_id,
-            posixpath.basename(path), parent_id)
+        ie = inventory.entry_factory[kind](file_id, name, parent_id)
         if kind == 'symlink':
             ie.symlink_target = self.get_symlink_target(file_id)
         else:
@@ -530,12 +589,13 @@ class GitWorkingTree(workingtree.WorkingTree):
     def stored_kind(self, file_id, path=None):
         if path is None:
             path = self.id2path(file_id)
-        head = self.repository._git.head()
-        if head == ZERO_SHA:
+        try:
+            return mode_kind(self.index[path.encode("utf-8")][4])
+        except KeyError:
+            # Maybe it's a directory?
+            if self._has_dir(path):
+                return "directory"
             raise errors.NoSuchId(self, file_id)
-        root_tree = self.store[head].tree
-        (mode, hexsha) = tree_lookup_path(self.store.__getitem__, root_tree, path)
-        return mode_kind(mode)
 
     if not osutils.supports_executable():
         def is_executable(self, file_id, path=None):
@@ -556,24 +616,48 @@ class GitWorkingTree(workingtree.WorkingTree):
 
     def list_files(self, include_root=False, from_dir=None, recursive=True):
         # FIXME: Yield non-versioned files
-        # FIXME: support from_dir
-        # FIXME: Support recursive
+        if from_dir is None:
+            from_dir = ""
         dir_ids = {}
+        fk_entries = {'directory': workingtree.TreeDirectory,
+                      'file': workingtree.TreeFile,
+                      'symlink': workingtree.TreeLink}
         root_ie = self._get_dir_ie(u"", None)
         if include_root and not from_dir:
             yield "", "V", root_ie.kind, root_ie.file_id, root_ie
         dir_ids[u""] = root_ie.file_id
-        for path, value in self.index.iteritems():
+        if recursive:
+            path_iterator = self._iter_files_recursive(from_dir)
+        else:
+            if from_dir is None:
+                start = self.basedir
+            else:
+                start = os.path.join(self.basedir, from_dir)
+            path_iterator = sorted([os.path.join(from_dir, name) for name in
+                os.listdir(start) if not self.bzrdir.is_control_filename(name)
+                and not self.mapping.is_special_file(name)])
+        for path in path_iterator:
+            try:
+                value = self.index[path]
+            except KeyError:
+                value = None
             path = path.decode("utf-8")
-            parent = posixpath.dirname(path).strip("/")
+            parent, name = posixpath.split(path)
             for dir_path, dir_ie in self._add_missing_parent_ids(parent, dir_ids):
                 yield dir_path, "V", dir_ie.kind, dir_ie.file_id, dir_ie
-            ie = self._get_file_ie(path, value, dir_ids[parent])
-            yield path, "V", ie.kind, ie.file_id, ie
+            if value is not None:
+                ie = self._get_file_ie(name, path, value, dir_ids[parent])
+                yield path, "V", ie.kind, ie.file_id, ie
+            else:
+                kind = osutils.file_kind(self.abspath(path))
+                ie = fk_entries[kind]()
+                yield path, "?", kind, None, ie
 
     def all_file_ids(self):
         ids = {u"": self.path2id("")}
         for path in self.index:
+            if self.mapping.is_special_file(path):
+                continue
             path = path.decode("utf-8")
             parent = posixpath.dirname(path).strip("/")
             for e in self._add_missing_parent_ids(parent, ids):
@@ -581,6 +665,10 @@ class GitWorkingTree(workingtree.WorkingTree):
             ids[path] = self.path2id(path)
         return set(ids.values())
 
+    def _directory_is_tree_reference(self, path):
+        # FIXME: Check .gitsubmodules for path
+        return False
+
     def iter_entries_by_dir(self, specific_file_ids=None, yield_parents=False):
         # FIXME: Is return order correct?
         if yield_parents:
@@ -598,14 +686,16 @@ class GitWorkingTree(workingtree.WorkingTree):
             yield u"", root_ie
         dir_ids = {u"": root_ie.file_id}
         for path, value in self.index.iteritems():
+            if self.mapping.is_special_file(path):
+                continue
             path = path.decode("utf-8")
             if specific_paths is not None and not path in specific_paths:
                 continue
+            (parent, name) = posixpath.split(path)
             try:
-                file_ie = self._get_file_ie(path, value, None)
+                file_ie = self._get_file_ie(name, path, value, None)
             except IOError:
                 continue
-            parent = posixpath.dirname(path).strip("/")
             for (dir_path, dir_ie) in self._add_missing_parent_ids(parent,
                     dir_ids):
                 yield dir_path, dir_ie
@@ -621,11 +711,20 @@ class GitWorkingTree(workingtree.WorkingTree):
         # The index just contains content, which won't have changed.
         self._reset_data()
 
+    def get_canonical_inventory_path(self, path):
+        for p in self.index:
+            if p.lower() == path.lower():
+                return p
+        else:
+            return path
+
     def _walkdirs(self, prefix=""):
         if prefix != "":
             prefix += "/"
         per_dir = defaultdict(list)
         for path, value in self.index.iteritems():
+            if self.mapping.is_special_file(path):
+                continue
             if not path.startswith(prefix):
                 continue
             (dirname, child_name) = posixpath.split(path)
@@ -633,7 +732,7 @@ class GitWorkingTree(workingtree.WorkingTree):
             dir_file_id = self.path2id(dirname)
             assert isinstance(value, tuple) and len(value) == 10
             (ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags) = value
-            stat_result = posix.stat_result((mode, ino,
+            stat_result = os.stat_result((mode, ino,
                     dev, 1, uid, gid, size,
                     0, mtime, ctime))
             per_dir[(dirname, dir_file_id)].append(
@@ -643,6 +742,7 @@ class GitWorkingTree(workingtree.WorkingTree):
                 mode_kind(mode)))
         return per_dir.iteritems()
 
+
 class GitWorkingTreeFormat(workingtree.WorkingTreeFormat):
 
     _tree_class = GitWorkingTree
@@ -687,21 +787,8 @@ class InterIndexGitTree(tree.InterTree):
         changes = self._index.changes_from_tree(
             self.source.store, self.source.tree, 
             want_unchanged=want_unchanged)
-        source_fileid_map = self.source.mapping.get_fileid_map(
-            self.source.store.__getitem__,
-            self.source.tree)
-        if self.target.mapping.BZR_FILE_IDS_FILE is not None:
-            file_id = self.target.path2id(
-                self.target.mapping.BZR_FILE_IDS_FILE)
-            if file_id is None:
-                target_fileid_map = {}
-            else:
-                target_fileid_map = self.target.mapping.import_fileid_map(
-                    Blob.from_string(self.target.get_file_text(file_id)))
-        else:
-            target_fileid_map = {}
-        target_fileid_map = GitFileIdMap(target_fileid_map,
-                self.target.mapping)
+        source_fileid_map = self.source._fileid_map
+        target_fileid_map = self.target._fileid_map
         ret = tree_delta_from_git_changes(changes, self.target.mapping,
             (source_fileid_map, target_fileid_map),
             specific_file=specific_files, require_versioned=require_versioned)
